from .combine import combine
from .combine_roulette import combine_roulette
from .maxent import maxent
from .mle import mle
from .quartile import quartile
from .quartile_int import QuartileInt
from .roulette import Roulette

__all__ = [
    "combine",
    "combine_roulette",
    "maxent",
    "mle",
    "quartile",
    "QuartileInt",
    "Roulette",
]



import pytest
from numpy.testing import assert_almost_equal
import numpy as np
from scipy import stats


from preliz import (
    AsymmetricLaplace,
    Beta,
    BetaScaled,
    Cauchy,
    ChiSquared,
    DiscreteUniform,
    ExGaussian,
    Exponential,
    Gamma,
    Gumbel,
    HalfCauchy,
    HalfNormal,
    HalfStudentT,
    InverseGamma,
    Kumaraswamy,
    Laplace,
    Logistic,
    LogLogistic,
    LogitNormal,
    LogNormal,
    Moyal,
    Normal,
    Pareto,
    Rice,
    SkewNormal,
    SkewStudentT,
    StudentT,
    Triangular,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
    Bernoulli,
    BetaBinomial,
    Binomial,
    Geometric,
    HyperGeometric,
    NegativeBinomial,
    Poisson,
    ZeroInflatedBinomial,
    ZeroInflatedNegativeBinomial,
    ZeroInflatedPoisson,
)


@pytest.mark.parametrize(
    "p_dist, sp_dist, p_params, sp_params",
    [
        (
            AsymmetricLaplace,
            stats.laplace_asymmetric,
            {"mu": 2.5, "b": 3.5, "kappa": 0.7},
            {"loc": 2.5, "scale": 3.5, "kappa": 0.7},
        ),
        (Beta, stats.beta, {"alpha": 2, "beta": 5}, {"a": 2, "b": 5}),
        (
            BetaScaled,
            stats.beta,
            {"alpha": 2, "beta": 5, "lower": -1, "upper": 3},
            {"a": 2, "b": 5, "loc": -1, "scale": 4},
        ),
        (Cauchy, stats.cauchy, {"alpha": 2, "beta": 4.5}, {"loc": 2, "scale": 4.5}),
        (ChiSquared, stats.chi2, {"nu": 3}, {"df": 3}),
        (
            ExGaussian,
            stats.exponnorm,
            {"mu": -1, "sigma": 2, "nu": 5},
            {"loc": -1, "scale": 2, "K": 5 / 2},
        ),
        (Exponential, stats.expon, {"beta": 3.7}, {"scale": 3.7}),
        (Gamma, stats.gamma, {"alpha": 2, "beta": 1 / 3}, {"a": 2, "scale": 3}),
        (Gumbel, stats.gumbel_r, {"mu": 2.5, "beta": 3.5}, {"loc": 2.5, "scale": 3.5}),
        (HalfCauchy, stats.halfcauchy, {"beta": 3.5}, {"scale": 3.5}),
        (HalfNormal, stats.halfnorm, {"sigma": 2}, {"scale": 2}),
        (
            HalfStudentT,
            stats.halfnorm,
            {"nu": 100, "sigma": 2},
            {"loc": 0, "scale": 2},
        ),  # not in scipy
        (InverseGamma, stats.invgamma, {"alpha": 5, "beta": 2}, {"a": 5, "scale": 2}),
        (
            Kumaraswamy,
            stats.beta,
            {"a": 1.00000001, "b": 5},
            {"a": 1.00000001, "b": 5},
        ),  # not in scipy
        (Laplace, stats.laplace, {"mu": 2.5, "b": 4}, {"loc": 2.5, "scale": 4}),
        (LogLogistic, stats.fisk, {"alpha": 1, "beta": 8}, {"c": 8}),
        (Logistic, stats.logistic, {"mu": 2.5, "s": 4}, {"loc": 2.5, "scale": 4}),
        (LogNormal, stats.lognorm, {"mu": 0, "sigma": 2}, {"s": 2, "scale": 1}),
        (LogitNormal, stats.beta, {"mu": 0, "sigma": 0.2}, {"a": 50.5, "b": 50.5}),  # not in scipy
        (Moyal, stats.moyal, {"mu": 1, "sigma": 2}, {"loc": 1, "scale": 2}),
        (Normal, stats.norm, {"mu": 0, "sigma": 2}, {"loc": 0, "scale": 2}),
        (Pareto, stats.pareto, {"m": 1, "alpha": 4.5}, {"b": 4.5}),
        (
            SkewNormal,
            stats.skewnorm,
            {"mu": 1, "sigma": 0.5, "alpha": 2},
            {"a": 2, "loc": 1, "scale": 0.5},
        ),
        (Rice, stats.rice, {"nu": 0.5, "sigma": 2}, {"b": 0.25, "scale": 2}),
        (
            SkewStudentT,
            stats.jf_skew_t,
            {"mu": 0.5, "sigma": 2, "a": 3, "b": 5},
            {"loc": 0.5, "scale": 2, "a": 3, "b": 5},
        ),
        (StudentT, stats.t, {"nu": 5, "mu": 0, "sigma": 2}, {"df": 5, "loc": 0, "scale": 2}),
        (Triangular, stats.triang, {"lower": 0, "upper": 1, "c": 0.45}, {"c": 0.45}),
        (
            TruncatedNormal,
            stats.truncnorm,
            {"mu": 0, "sigma": 1, "lower": -1, "upper": 1},
            {"loc": 0, "scale": 1, "a": -1, "b": 1},
        ),
        (Uniform, stats.uniform, {"lower": -2, "upper": 1}, {"loc": -2, "scale": 3}),
        (VonMises, stats.vonmises, {"mu": 0, "kappa": 10}, {"loc": 0, "kappa": 10}),
        (Wald, stats.invgauss, {"mu": 2, "lam": 10}, {"mu": 2 / 10, "scale": 10}),
        (
            Weibull,
            stats.weibull_min,
            {"alpha": 5.0, "beta": 2.0},
            {"c": 5.0, "scale": 2.0},
        ),
        (Binomial, stats.binom, {"n": 4, "p": 0.4}, {"n": 4, "p": 0.4}),
        (
            BetaBinomial,
            stats.betabinom,
            {"alpha": 2, "beta": 5, "n": 10},
            {"n": 10, "a": 2, "b": 5},
        ),
        (Bernoulli, stats.bernoulli, {"p": 0.4}, {"p": 0.4}),
        (DiscreteUniform, stats.randint, {"lower": -2, "upper": 1}, {"low": -2, "high": 2}),
        (Geometric, stats.geom, {"p": 0.4}, {"p": 0.4}),
        (HyperGeometric, stats.hypergeom, {"N": 50, "k": 10, "n": 25}, {"M": 50, "N": 25, "n": 10}),
        (
            NegativeBinomial,
            stats.nbinom,
            {"mu": 3.5, "alpha": 2.1},
            {"n": 2.1, "p": 2.1 / (3.5 + 2.1)},
        ),
        (Poisson, stats.poisson, {"mu": 3.5}, {"mu": 3.5}),
        (
            ZeroInflatedBinomial,  # not in scipy
            stats.binom,
            {"psi": 1, "n": 4, "p": 0.4},
            {"n": 4, "p": 0.4},
        ),
        (
            ZeroInflatedNegativeBinomial,  # not in scipy
            stats.nbinom,
            {"psi": 1, "mu": 3.5, "alpha": 2.1},
            {"n": 2.1, "p": 2.1 / (3.5 + 2.1)},
        ),
        (
            ZeroInflatedPoisson,  # not in scipy
            stats.poisson,
            {"psi": 1, "mu": 3.5},
            {"mu": 3.5},
        ),
    ],
)
def test_match_scipy(p_dist, sp_dist, p_params, sp_params):
    preliz_dist = p_dist(**p_params)
    scipy_dist = sp_dist(**sp_params)
    preliz_name = preliz_dist.__class__.__name__

    if preliz_name != "VonMises":
        # for the VonMises we used the differential entropy definition.
        # SciPy uses a different one
        actual = preliz_dist.entropy()
        expected = scipy_dist.entropy()
        if preliz_dist.kind == "discrete":
            assert_almost_equal(actual, expected, decimal=1)
        elif preliz_name in [
            "HalfStudentT",
            "Moyal",
            "LogitNormal",
            "SkewNormal",
            "SkewStudentT",
            "Rice",
            "ExGaussian",
        ]:
            assert_almost_equal(actual, expected, decimal=2)
        else:
            assert_almost_equal(actual, expected, decimal=4)

    rng = np.random.default_rng(1)
    actual_rvs = preliz_dist.rvs(20, random_state=rng)
    rng = np.random.default_rng(1)
    expected_rvs = scipy_dist.rvs(20, random_state=rng)
    if preliz_name in [
        "ExGaussian",
        "HalfStudentT",
        "Kumaraswamy",
        "LogitNormal",
        "Moyal",
        "StudentT",
        "Weibull",
        "InverseGamma",
        "DiscreteUniform",
        "ZeroInflatedBinomial",
        "ZeroInflatedNegativeBinomial",
        "ZeroInflatedPoisson",
    ]:
        pz_rvs = preliz_dist.rvs(20000, random_state=rng)
        sc_rvs = scipy_dist.rvs(20000, random_state=rng)
        assert_almost_equal(pz_rvs.mean(), sc_rvs.mean(), decimal=1)
        assert_almost_equal(pz_rvs.std(), sc_rvs.std(), decimal=1)

    else:
        assert_almost_equal(actual_rvs, expected_rvs)

    support = preliz_dist.support
    if preliz_name == "VonMises":
        extended_vals = actual_rvs
    else:
        extended_vals = np.concatenate(
            [
                actual_rvs,
                support,
                [support[0] - 1],
                [support[0] - 2],
                [support[1] + 1],
                [support[1] + 2],
            ]
        )

    actual_pdf = preliz_dist.pdf(extended_vals)
    if preliz_dist.kind == "continuous":
        expected_pdf = scipy_dist.pdf(extended_vals)
    else:
        expected_pdf = scipy_dist.pmf(extended_vals)

    if preliz_name == "LogitNormal":
        assert_almost_equal(actual_pdf, expected_pdf, decimal=1)
    elif preliz_name == "HalfStudentT":
        assert_almost_equal(actual_pdf, expected_pdf, decimal=2)
    else:
        assert_almost_equal(actual_pdf, expected_pdf, decimal=4)

    actual_cdf = preliz_dist.cdf(extended_vals)
    expected_cdf = scipy_dist.cdf(extended_vals)

    if preliz_name in ["HalfStudentT", "LogitNormal"]:
        assert_almost_equal(actual_cdf, expected_cdf, decimal=2)
    else:
        assert_almost_equal(actual_cdf, expected_cdf, decimal=6)

    x_vals = [-1, 0, 0.25, 0.5, 0.75, 1, 2]
    actual_ppf = preliz_dist.ppf(x_vals)
    expected_ppf = scipy_dist.ppf(x_vals)
    if preliz_name in ["HalfStudentT", "Wald", "LogitNormal", "SkewNormal", "ExGaussian"]:
        assert_almost_equal(actual_ppf, expected_ppf, decimal=2)
    else:
        assert_almost_equal(actual_ppf, expected_ppf)

    actual_logpdf = preliz_dist.logpdf(extended_vals)
    if preliz_dist.kind == "continuous":
        expected_logpdf = scipy_dist.logpdf(extended_vals)
    else:
        expected_logpdf = scipy_dist.logpmf(extended_vals)

    if preliz_name == "HalfStudentT":
        assert_almost_equal(actual_logpdf, expected_logpdf, decimal=0)
    elif preliz_name == "LogitNormal":
        assert_almost_equal(actual_logpdf, expected_logpdf, decimal=1)
    else:
        assert_almost_equal(actual_logpdf, expected_logpdf)

    actual_neg_logpdf = preliz_dist._neg_logpdf(extended_vals)
    expected_neg_logpdf = -expected_logpdf.sum()
    if preliz_name in ["HalfStudentT", "LogitNormal"]:
        assert_almost_equal(actual_neg_logpdf, expected_neg_logpdf, decimal=1)
    elif preliz_name in ["TruncatedNormal"]:
        assert_almost_equal(actual_neg_logpdf, expected_neg_logpdf, decimal=6)
    else:
        assert_almost_equal(actual_neg_logpdf, expected_neg_logpdf)

    if preliz_dist.__class__.__name__ not in [
        "HalfStudentT",
        "LogLogistic",
        "Rice",
        "VonMises",
        "ZeroInflatedBinomial",
        "ZeroInflatedNegativeBinomial",
        "ZeroInflatedPoisson",
    ]:
        actual_moments = preliz_dist.moments("mvsk")
        expected_moments = scipy_dist.stats("mvsk")
    elif preliz_dist.__class__.__name__ == "VonMises":
        # We use the circular variance definition for the variance
        assert_almost_equal(preliz_dist.var(), stats.circvar(preliz_dist.rvs(1000)), decimal=1)
        # And we adopt the convention of setting the skewness and kurtosis to 0 in
        # analogy with the Normal distribution
        actual_moments = preliz_dist.moments("m")
        expected_moments = scipy_dist.stats("m")

    else:
        actual_moments = preliz_dist.moments("mv")
        expected_moments = scipy_dist.stats("mv")

    if preliz_name in ["HalfStudentT", "LogitNormal"]:
        assert_almost_equal(actual_moments, expected_moments, decimal=1)
    elif preliz_name in ["TruncatedNormal"]:
        assert_almost_equal(actual_moments, expected_moments, decimal=6)
    else:
        assert_almost_equal(actual_moments, expected_moments)

    actual_median = preliz_dist.median()
    expected_median = scipy_dist.median()

    if preliz_name == "HalfStudentT":
        assert_almost_equal(actual_median, expected_median, decimal=1)
    elif preliz_name in ["SkewNormal", "ExGaussian"]:
        assert_almost_equal(actual_median, expected_median, decimal=6)
    else:
        assert_almost_equal(actual_median, expected_median)

    finite_expected_pdf = np.where(np.isfinite(expected_pdf), expected_pdf, -np.inf)
    expected_mode = extended_vals[np.argmax(finite_expected_pdf)]
    try:
        actual_mode = preliz_dist.mode()
        assert_almost_equal(actual_mode, expected_mode, decimal=0)
    except NotImplementedError:
        pass



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import eps, all_not_none
from ..internal.optimization import optimize_ml
from ..internal.special import (
    garcia_approximation,
    gamma,
    mean_and_std,
    cdf_bounds,
    ppf_bounds_cont,
    xlogy,
)


class Weibull(Continuous):
    r"""
    Weibull distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \alpha, \beta) =
           \frac{\alpha x^{\alpha - 1}
           \exp(-(\frac{x}{\beta})^{\alpha})}{\beta^\alpha}

    .. plot::
        :context: close-figs


        from preliz import Weibull, style
        style.use('preliz-doc')
        alphas = [1., 2, 5.]
        betas = [1., 1., 2.]
        for a, b in zip(alphas, betas):
            Weibull(a, b).plot_pdf(support=(0,5))

    ========  ====================================================
    Support   :math:`x \in [0, \infty)`
    Mean      :math:`\beta \Gamma(1 + \frac{1}{\alpha})`
    Variance  :math:`\beta^2 \Gamma(1 + \frac{2}{\alpha} - \mu^2/\beta^2)`
    ========  ====================================================

    Parameters
    ----------
    alpha : float
        Shape parameter (alpha > 0).
    beta : float
        Scale parameter (beta > 0).
    """

    def __init__(self, alpha=None, beta=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(alpha, beta)

    def _parametrization(self, alpha=None, beta=None):
        self.alpha = alpha
        self.beta = beta
        self.param_names = ("alpha", "beta")
        self.params_support = ((eps, np.inf), (eps, np.inf))
        if all_not_none(alpha, beta):
            self._update(alpha, beta)

    def _update(self, alpha, beta):
        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.params = (self.alpha, self.beta)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.alpha, self.beta))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.beta, self.support[0], self.support[1])

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.beta, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta)

    def entropy(self):
        return nb_entropy(self.alpha, self.beta)

    def mean(self):
        return self.beta * gamma(1 + 1 / self.alpha)

    def median(self):
        return self.beta * np.log(2) ** (1 / self.alpha)

    def var(self):
        return self.beta**2 * gamma(1 + 2 / self.alpha) - self.mean() ** 2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        mu = self.mean()
        sigma = self.std()
        m_s = mu / sigma
        return gamma(1 + 3 / self.alpha) * (self.beta / sigma) ** 3 - 3 * m_s - m_s**3

    def kurtosis(self):
        mu = self.mean()
        sigma = self.std()
        skew = self.skewness()
        m_s = mu / sigma
        return (
            (self.beta / sigma) ** 4 * gamma(1 + 4 / self.alpha)
            - 4 * skew * m_s
            - 6 * m_s**2
            - m_s**4
            - 3
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.weibull(self.alpha, size) * self.beta

    def _fit_moments(self, mean, sigma):
        alpha, beta = garcia_approximation(mean, sigma)
        self._update(alpha, beta)

    def _fit_mle(self, sample):
        mean, std = mean_and_std(sample)
        self._fit_moments(mean, std)
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, alpha, beta, lower, upper):
    prob = 1 - np.exp(-((x / beta) ** alpha))
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, alpha, beta, lower, upper):
    x_val = beta * (-np.log(1 - q)) ** (1 / alpha)
    return ppf_bounds_cont(x_val, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(alpha, beta):
    return np.euler_gamma * (1 - 1 / alpha) + np.log(beta / alpha) + 1


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, beta):
    if x < 0:
        return -np.inf
    else:
        x_b = x / beta
        return np.log(alpha / beta) + xlogy((alpha - 1), x_b) - x_b**alpha


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta):
    return -(nb_logpdf(x, alpha, beta)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
# pylint: disable=invalid-name
import numpy as np
import numba as nb

from ..internal.distribution_helper import all_not_none
from .distributions import Continuous


class Triangular(Continuous):
    r"""
    Triangular distribution

    The pdf of this distribution is

    .. math::

        \begin{cases}
            0 & \text{for } x < a, \\
            \frac{2(x-a)}{(b-a)(c-a)} & \text{for } a \le x < c, \\[4pt]
            \frac{2}{b-a}             & \text{for } x = c, \\[4pt]
            \frac{2(b-x)}{(b-a)(b-c)} & \text{for } c < x \le b, \\[4pt]
            0 & \text{for } b < x.
        \end{cases}

    .. plot::
        :context: close-figs

        
        from preliz import Triangular, style
        style.use('preliz-doc')
        lowers = [0., -1, 2]
        cs = [2., 0., 6.5]
        uppers = [4., 1, 8]
        for lower, c, upper in zip(lowers, cs, uppers):
            scale = upper - lower
            c_ = (c - lower) / scale
            Triangular(lower, c, upper).plot_pdf()

    ========  ============================================================================
    Support   :math:`x \in [lower, upper]`
    Mean      :math:`\dfrac{lower + upper + c}{3}`
    Variance  :math:`\dfrac{upper^2 + lower^2 +c^2 - lower*upper - lower*c - upper*c}{18}`
    ========  ============================================================================

    Parameters
    ----------
    lower : float
        Lower limit.
    c : float
        Mode.
    upper : float
        Upper limit.
    """

    def __init__(self, lower=None, c=None, upper=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(lower, c, upper)

    def _parametrization(self, lower=None, c=None, upper=None):
        self.lower = lower
        self.c = c
        self.upper = upper
        self.params = (self.lower, self.c, self.upper)
        self.param_names = ("lower", "c", "upper")
        self.params_support = ((-np.inf, np.inf), (-np.inf, np.inf), (-np.inf, np.inf))
        if all_not_none(lower, c, upper):
            self._update(lower, c, upper)

    def _update(self, lower, c, upper):
        self.lower = np.float64(lower)
        self.c = np.float64(c)
        self.upper = np.float64(upper)
        self.support = (self.lower, self.upper)
        self.params = (self.lower, self.c, self.upper)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.lower, self.c, self.upper))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.lower, self.c, self.upper)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.lower, self.c, self.upper)

    def logpdf(self, x):
        return nb_logpdf(x, self.lower, self.c, self.upper)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.lower, self.c, self.upper)

    def entropy(self):
        return nb_entropy(self.lower, self.upper)

    def mean(self):
        return (self.lower + self.c + self.upper) / 3

    def mode(self):
        return self.c

    def median(self):
        return np.where(
            self.c >= (self.lower + self.upper) / 2,
            self.lower + ((self.upper - self.lower) * (self.c - self.lower) / 2) ** 0.5,
            self.upper - ((self.upper - self.lower) * (self.upper - self.c) / 2) ** 0.5,
        )

    def var(self):
        return (
            self.lower**2
            + self.upper**2
            + self.c**2
            - self.lower * self.c
            - self.c * self.upper
            - self.lower * self.upper
        ) / 18

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (
            2**0.5
            * (self.lower + self.upper - 2 * self.c)
            * (2 * self.lower - self.upper - self.c)
            * (self.lower - 2 * self.upper + self.c)
        ) / (
            5
            * (
                self.lower**2
                + self.upper**2
                + self.c**2
                - self.lower * self.c
                - self.c * self.upper
                - self.lower * self.upper
            )
            ** (3 / 2)
        )

    def kurtosis(self):
        return -3 / 5

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        random_samples = random_state.uniform(0, 1, size)
        return nb_rvs(random_samples, self.lower, self.c, self.upper)

    def _fit_moments(self, mean, sigma):
        # Assume symmetry
        lower = mean - 6**0.5 * sigma
        upper = mean + 6**0.5 * sigma
        c = mean
        self._update(lower, c, upper)

    def _fit_mle(self, sample):
        lower = np.min(sample)
        upper = np.max(sample)
        middle = (np.mean(sample) * 3) - lower - upper
        self._update(lower, middle, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_cdf(x, lower, c, upper):
    if x <= lower:
        return 0
    elif lower < x <= c:
        return (x - lower) ** 2 / ((upper - lower) * (c - lower))
    elif c < x < upper:
        return 1 - (upper - x) ** 2 / ((upper - lower) * (upper - c))
    return 1


@nb.vectorize(nopython=True, cache=True)
def nb_ppf(q, lower, c, upper):
    if 0 <= q < (c - lower) / (upper - lower):
        return lower + (q * (upper - lower) * (c - lower)) ** 0.5
    elif (c - lower) / (upper - lower) <= q <= 1:
        return upper - ((1 - q) * (upper - lower) * (upper - c)) ** 0.5
    return np.nan


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, lower, c, upper):
    if x < lower:
        return -np.inf
    elif lower <= x < c:
        return np.log(2 * (x - lower) / ((upper - lower) * (c - lower)))
    elif x == c:
        return np.log(2 / (upper - lower))
    elif c < x <= upper:
        return np.log(2 * (upper - x) / ((upper - lower) * (upper - c)))
    return -np.inf


@nb.njit(cache=True)
def nb_neg_logpdf(x, lower, c, upper):
    return -(nb_logpdf(x, lower, c, upper)).sum()


@nb.njit(cache=True)
def nb_entropy(lower, upper):
    return 0.5 + np.log((upper - lower) / 2)


@nb.vectorize(nopython=True, cache=True)
def nb_rvs(random_samples, lower, c, upper):
    if 0 < random_samples < (c - lower) / (upper - lower):
        return lower + (random_samples * (upper - lower) * (c - lower)) ** 0.5
    return upper - ((1 - random_samples) * (upper - lower) * (upper - c)) ** 0.5



from numpy.testing import assert_almost_equal
import numpy as np

from preliz.distributions import Truncated, TruncatedNormal, Normal


def test_truncated():
    custom_truncnorm_dist = Truncated(Normal(0, 2), -1, np.inf)
    genera_truncnorm_dist = TruncatedNormal(0, 2, -1, np.inf)

    rng = np.random.default_rng(1)
    actual_rvs = custom_truncnorm_dist.rvs(20, random_state=rng)
    rng = np.random.default_rng(1)
    expected_rvs = genera_truncnorm_dist.rvs(20, random_state=rng)
    assert_almost_equal(actual_rvs, expected_rvs)

    actual_pdf = custom_truncnorm_dist.pdf(actual_rvs)
    expected_pdf = genera_truncnorm_dist.pdf(actual_rvs)
    assert_almost_equal(actual_pdf, expected_pdf, decimal=4)

    support = custom_truncnorm_dist.support
    cdf_vals = np.concatenate([actual_rvs, support, [support[0] - 1], [support[1] + 1]])

    actual_cdf = custom_truncnorm_dist.cdf(cdf_vals)
    expected_cdf = genera_truncnorm_dist.cdf(cdf_vals)
    assert_almost_equal(actual_cdf, expected_cdf, decimal=6)

    x_vals = [-1, 0, 0.25, 0.5, 0.75, 1, 2]
    actual_ppf = custom_truncnorm_dist.ppf(x_vals)
    expected_ppf = genera_truncnorm_dist.ppf(x_vals)
    assert_almost_equal(actual_ppf, expected_ppf)

    actual_logpdf = custom_truncnorm_dist.logpdf(actual_rvs)
    expected_logpdf = genera_truncnorm_dist.logpdf(actual_rvs)
    assert_almost_equal(actual_logpdf, expected_logpdf)

    actual_neg_logpdf = custom_truncnorm_dist._neg_logpdf(actual_rvs)
    expected_neg_logpdf = -expected_logpdf.sum()
    assert_almost_equal(actual_neg_logpdf, expected_neg_logpdf)

    actual_median = custom_truncnorm_dist.median()
    expected_median = genera_truncnorm_dist.median()
    assert_almost_equal(actual_median, expected_median)

    actual_mean = custom_truncnorm_dist.mean()
    expected_mean = genera_truncnorm_dist.mean()
    assert_almost_equal(actual_mean, expected_mean, decimal=2)

    actual_var = custom_truncnorm_dist.var()
    expected_var = genera_truncnorm_dist.var()
    assert_almost_equal(actual_var, expected_var, decimal=2)

    actual_skew = custom_truncnorm_dist.skewness()
    expected_skew = genera_truncnorm_dist.skewness()
    assert_almost_equal(actual_skew, expected_skew, decimal=2)

    actual_kurt = custom_truncnorm_dist.kurtosis()
    expected_kurt = genera_truncnorm_dist.kurtosis()
    assert_almost_equal(actual_kurt, expected_kurt, decimal=1)

    actual_entropy = custom_truncnorm_dist.entropy()
    expected_entropy = genera_truncnorm_dist.entropy()
    assert_almost_equal(actual_entropy, expected_entropy, decimal=2)



"""
Parent classes for all families.
"""
# pylint: disable=no-member
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
from collections import namedtuple
from copy import copy
import warnings

try:
    from ipywidgets import interactive
except ImportError:
    pass
import numpy as np

from ..internal.plot_helper import (
    plot_pdfpmf,
    plot_cdf,
    plot_ppf,
    get_slider,
    check_inside_notebook,
)
from ..internal.distribution_helper import (
    init_vals,
    valid_scalar_params,
    valid_distribution,
)

from ..internal.optimization import optimize_hdi
from ..internal.rcparams import rcParams


class Distribution:
    """
    Base class for distributions.

    Not intended for direct instantiation.
    """

    def __init__(self):
        self.is_frozen = False
        self.opt = None

    def __repr__(self):
        name = self.__class__.__name__
        if name in ["Truncated", "Censored", "Hurdle"]:
            name += self.dist.__class__.__name__
        if name == "Mixture":
            name = (
                "Mixture"
                + "".join(dict.fromkeys(dist.__class__.__name__ for dist in self.dist))
                + "\n"
            )

        if self.is_frozen:
            if "Mixture" in name:
                bolded_name = "\033[1m" + name.strip() + "\033[0m" + "\n"
            else:
                bolded_name = "\033[1m" + name + "\033[0m"

            description = "".join(
                f"{n}={v:.3g}, "
                if np.isscalar(v)
                else f"{n}=[" + "".join(f"{vi:.3g}, " for vi in v).strip(", ") + "], "
                for n, v in zip(self.param_names, self.params)
            ).strip(", ")
            return f"{bolded_name}({description})"
        else:
            return name

    @property
    def params_dict(self):
        if self.is_frozen:
            return dict(zip(self.param_names, self.params))
        else:
            return None

    def summary(self, mass=None, interval=None, fmt=".2f"):
        """
        Namedtuple with the mean, median, standard deviation, and lower and upper bounds
        of the equal-tailed interval.

        Parameters
        ----------
        mass: float
            Probability mass for the equal-tailed interval. Defaults to None,
        which results in the value of rcParams["stats.ci_prob"] being used.
        interval : str or list-like
            Type of interval. Available options are highest density interval `"hdi"`,
            equal tailed interval `"eti"` or arbitrary interval defined by a list-like object
            with a pair of values. Defaults to the value in rcParams["stats.ci_kind"].
        fmt : str
            fmt used to represent results using f-string fmt for floats. Default to ".2f"
            i.e. 2 digits after the decimal point.
        """
        valid_distribution(self)

        if interval is None:
            interval = rcParams["stats.ci_kind"]

        if not isinstance(fmt, str):
            raise ValueError("Invalid format string.")

        if valid_scalar_params(self):
            name = self.__class__.__name__
            if name == "Truncated":
                name = "Truncated" + self.dist.__class__.__name__
            elif name == "Censored":
                name = "Censored" + self.dist.__class__.__name__
            elif name == "Mixture":
                name = "Mixture" + "".join(
                    dict.fromkeys(dist.__class__.__name__ for dist in self.dist)
                )

            attr = namedtuple(name, ["mean", "median", "std", "lower", "upper"])
            mean = float(f"{self.mean():{fmt}}")
            median = float(f"{self.median():{fmt}}")
            std = float(f"{self.std():{fmt}}")

            if isinstance(interval, (tuple, list, np.ndarray)):
                c_int = self.ppf(interval)
            elif interval == "hdi":
                c_int = self.hdi(mass, fmt=fmt)
            elif interval == "eti":
                c_int = self.eti(mass, fmt=fmt)

            if self.kind == "discrete":
                lower_tail = c_int[0]
                upper_tail = c_int[1]
            else:
                lower_tail = float(f"{c_int[0]:{fmt}}")
                upper_tail = float(f"{c_int[1]:{fmt}}")
            return attr(mean, median, std, lower_tail, upper_tail)
        else:
            return None

    def rvs(self, size=None, random_state=None):
        """Random sample

        Parameters
        ----------
        size : int or tuple of ints, optional
            Defining number of random variates. Defaults to 1.
        random_state : {None, int, numpy.random.Generator, numpy.random.RandomState}
            Defaults to None
        """
        raise NotImplementedError

    def pdf(self, x):
        """Probability density/mass function.

        Parameters
        ----------
        x : array_like
            Values on which to evaluate the pdf
        """
        raise NotImplementedError

    def cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : array_like
            Values on which to evaluate the cdf
        """
        raise NotImplementedError

    def ppf(self, q):
        """Percent point function (inverse of cdf).

        Parameters
        ----------
        x : array_like
            Values on which to evaluate the inverse of the cdf
        """
        raise NotImplementedError

    def logpdf(self, x):
        """Log probability density/mass function.

        Parameters
        ----------
        x : array_like
            Values on which to evaluate the logpdf
        """
        raise NotImplementedError

    def entropy(self):
        """Entropy"""
        raise NotImplementedError

    def mean(self):
        """Mean."""
        raise NotImplementedError

    def median(self):
        """Median."""
        raise NotImplementedError

    def mode(self):
        """Mode."""
        raise NotImplementedError

    def std(self):
        """Standard deviation."""
        raise NotImplementedError

    def var(self):
        """Variance"""
        raise NotImplementedError

    def skewness(self):
        """Skewness."""
        raise NotImplementedError

    def kurtosis(self):
        """Excess kurtosis."""
        raise NotImplementedError

    def moments(self, types="mvsk"):
        """
        Compute moments of the distribution.

        It can also return the standard deviation

        Parameters
        ----------
        types : str
            The type of moments to compute. Default is 'mvsk'
            where 'm' = mean, 'v' = variance, 's' = skewness, and 'k' = kurtosis.
            To compute the standard deviation use 'd'
            Valid combinations are any subset of 'mvdsk'.
        """
        moments = []
        for m_t in types:
            if m_t not in "mdvsk":
                raise ValueError(
                    "The input string should only contain the letters "
                    "'m', 'd', 'v', 's', or 'k'."
                )
            if m_t == "m":
                moments.append(self.mean())
            elif m_t == "d":
                moments.append(self.std())
            elif m_t == "v":
                moments.append(self.var())
            elif m_t == "s":
                moments.append(self.skewness())
            elif m_t == "k":
                moments.append(self.kurtosis())

        return moments

    def eti(self, mass=None, fmt=".2f"):
        """Equal-tailed interval containing `mass`.

        Parameters
        ----------
        mass: float
            Probability mass in the interval. Defaults to None,
        which results in the value of rcParams["stats.ci_prob"] being used.
        fmt : str
            fmt used to represent results using f-string fmt for floats. Default to ".2f"
            i.e. 2 digits after the decimal point. Use `"none"` for no format.
        """
        valid_distribution(self)

        if mass is None:
            mass = rcParams["stats.ci_prob"]

        if not isinstance(fmt, str):
            raise ValueError("Invalid format string.")

        if valid_scalar_params(self):
            lower_tail, upper_tail = self.ppf([(1 - mass) / 2, 1 - (1 - mass) / 2])
            if self.kind == "continuous" and fmt != "none":
                lower_tail = float(f"{lower_tail:{fmt}}")
                upper_tail = float(f"{upper_tail:{fmt}}")
            elif self.kind == "discrete":
                lower_tail = int(lower_tail)
                upper_tail = int(upper_tail)

            return (lower_tail, upper_tail)
        else:
            return None

    def hdi(self, mass=None, fmt=".2f"):
        """Highest density interval containing `mass`.

        Parameters
        ----------
        mass: float
            Probability mass in the interval. Defaults to None,
        which results in the value of rcParams["stats.ci_prob"] being used.
        fmt : str
            fmt used to represent results using f-string fmt for floats. Default to ".2f"
            i.e. 2 digits after the decimal point. Use `"none"` for no format.
        """
        valid_distribution(self)

        if mass is None:
            mass = rcParams["stats.ci_prob"]

        if not isinstance(fmt, str):
            raise ValueError("Invalid format string.")

        if self.__class__.__name__ == "Mixture":
            warnings.warn("HDI may not be correct for multimodal distributions")

        if valid_scalar_params(self):
            lower_tail, upper_tail = optimize_hdi(self, mass)
            if self.kind == "continuous" and fmt != "none":
                lower_tail = float(f"{lower_tail:{fmt}}")
                upper_tail = float(f"{upper_tail:{fmt}}")
            return (lower_tail, upper_tail)
        else:
            return None

    def to_pymc(self, name=None, **kwargs):
        """
        Convert the PreliZ distribution to a PyMC distribution.

        name : str
            Name of PyMC distribution. Needed if inside Model context
        kwargs : PyMC distributions properties
            kwargs are used to specify properties such as shape or dims

        Returns
        -------
        PyMC distribution
        """
        try:
            import pymc.distributions as pm_dists
            from pymc.model import Model

            model = Model.get_context(error_if_none=False)

            if self.__class__.__name__ == "Hurdle":
                preliz_name = self.__class__.__name__ + self.dist.__class__.__name__
            else:
                preliz_name = self.__class__.__name__
            pymc_class = getattr(pm_dists, preliz_name)

            if model is None:
                if self.__class__.__name__ in ["Truncated", "Censored"]:
                    pymc_dist = pymc_class.dist(
                        self.dist.to_pymc(),
                        lower=self.params_dict["lower"],
                        upper=self.params_dict["upper"],
                        **kwargs,
                    )
                elif self.__class__.__name__ == "Mixture":
                    pymc_dist = pymc_class.dist(
                        self.weights,
                        [dist.to_pymc() for dist in self.dist],
                        **kwargs,
                    )
                else:
                    pymc_dist = pymc_class.dist(**self.params_dict, **kwargs)
            else:
                if self.__class__.__name__ in ["Truncated", "Censored"]:
                    pymc_dist = pymc_class(
                        name,
                        getattr(pm_dists, self.dist.__class__.__name__).dist(
                            **self.dist.params_dict
                        ),
                        lower=self.params_dict["lower"],
                        upper=self.params_dict["upper"],
                        **kwargs,
                    )
                elif self.__class__.__name__ == "Mixture":
                    pymc_dist = pymc_class(
                        name,
                        self.weights,
                        [
                            getattr(pm_dists, dist.__class__.__name__).dist(**dist.params_dict)
                            for dist in self.dist
                        ],
                        **kwargs,
                    )
                else:
                    pymc_dist = pymc_class(name, **self.params_dict, **kwargs)

            return pymc_dist

        except ImportError:
            raise ImportError("This function requires PyMC") from None

    def to_bambi(self, **kwargs):
        """
        Convert the PreliZ distribution to a Bambi Prior.

        kwargs : PyMC distributions properties
            kwargs are used to specify properties such as shape or dims

        Returns
        -------
        Bambi Prior
        """
        try:
            from bambi import Prior

            return Prior(self.__class__.__name__, **self.params_dict, **kwargs)

        except ImportError:
            raise ImportError("This function requires Bambi") from None

    def _check_endpoints(self, lower, upper, raise_error=True):
        """
        Evaluate if the lower and upper values are in the support of the distribution

        Parameters
        ----------
        lower : int or float
            lower endpoint
        upper : int or float
            upper endpoint
        raise_error : bool
            If True (default) it will raise ValueErrors, otherwise it will return True
            if the lower and upper endpoint are in the support of the distribution or
            False otherwise.
        """
        s_l, s_u = self.support

        if raise_error:
            domain_error = (
                f"The provided endpoints are outside the domain of the "
                f"{self.__class__.__name__} distribution"
            )

            if np.isfinite(s_l):
                if lower < s_l:
                    raise ValueError(domain_error)

            if np.isfinite(s_u):
                if upper > s_u:
                    raise ValueError(domain_error)
            if np.isfinite(s_l) and np.isfinite(s_u):
                if lower == s_l and upper == s_u:
                    raise ValueError(
                        "Given the provided endpoints, mass will be always 1. "
                        "Please provide other values"
                    )
            return None
        else:
            return lower >= s_l and upper <= s_u

    def _finite_endpoints(self, support):
        """
        Return finite endpoints even for unbounded distributions

        Parameters
        ----------
        support : str or tuple
            Available string options are "full" or "restricted".
        """
        if isinstance(support, tuple):
            lower_ep, upper_ep = support
        else:
            if support not in ["restricted", "full"]:
                raise ValueError("Allowed values for the support are 'restricted' or 'full' ")

            lower_ep, upper_ep = self.support

            if not np.isfinite(lower_ep) or support == "restricted":
                lower_ep = self.ppf(0.0001)
            if not np.isfinite(upper_ep) or support == "restricted":
                upper_ep = self.ppf(0.9999)

        return lower_ep, upper_ep

    def xvals(self, support, n_points=None):
        """Provide x values in the support of the distribution. This is useful for example when
        plotting.

        Parameters
        ----------
        support : str
            Available options are `"full"` or `"restricted"`.
            If `"full"` the values will cover the entire support of the distribution if the boundary
            is finite, or the quantiles 0.0001 or 0.9999, if infinite.
            If `"restricted"` the values will cover the quantile 0.0001 to 0.9999.
        n_points : int
            Number of values to return. Defaults to 1000 for continuous distributions
            and 200 for discrete ones.
            For discrete distributions the returned values may be fewer
            than `n_points` if the actual number of discrete values in the support of the
            distribution is smaller than `n_points`.
        """
        lower_ep, upper_ep = self._finite_endpoints(support)

        if self.kind == "continuous":
            if n_points is None:
                n_points = 1000
            return _continuous_xvals(lower_ep, upper_ep, n_points)
        else:
            if n_points is None:
                n_points = 200
            return _discrete_xvals(lower_ep, upper_ep, n_points)

    def plot_pdf(
        self,
        moments=None,
        pointinterval=False,
        interval=None,
        levels=None,
        support="restricted",
        legend="legend",
        color=None,
        alpha=1,
        figsize=None,
        ax=None,
    ):
        """
        Plot the  pdf (continuous) or pmf (discrete).

        Parameters
        ----------
        moments : str
            Compute moments. Use any combination of the strings ``m``, ``d``, ``v``, ``s`` or ``k``
            for the mean (μ), standard deviation (σ), variance (σ²), skew (γ) or kurtosis (κ)
            respectively. Other strings will be ignored. Defaults to None.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"` (default),
            equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
            Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"legend"`` a
            ``"title"`` or not include them ``None``.
        color : str
            Valid matplotlib color. Defaults to None, colors will be selected from the active
            color cycle.
        alpha : float
            Transparency of the line. Defaults to 1 (no transparency).
        figsize : tuple
            Size of the figure
        ax : matplotlib axes
        """

        if valid_scalar_params(self):
            return plot_pdfpmf(
                self,
                moments,
                pointinterval,
                interval,
                levels,
                support,
                legend,
                color,
                alpha,
                figsize,
                ax,
            )
        else:
            return None

    def plot_cdf(
        self,
        moments=None,
        pointinterval=False,
        interval=None,
        levels=None,
        support="restricted",
        legend="legend",
        color=None,
        alpha=1,
        figsize=None,
        ax=None,
    ):
        """
        Plot the cumulative distribution function.

        Parameters
        ----------
        moments : str
            Compute moments. Use any combination of the strings ``m``, ``d``, ``v``, ``s`` or ``k``
            for the mean (μ), standard deviation (σ), variance (σ²), skew (γ) or kurtosis (κ)
            respectively. Other strings will be ignored. Defaults to None.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"` (default),
            equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
            Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        color : str
            Valid matplotlib color. Defaults to None, colors will be selected from the active
            color cycle.
        alpha : float
            Transparency of the line. Defaults to 1 (no transparency).
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"legend"`` a
            ``"title"`` or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axes
        """
        if valid_scalar_params(self):
            return plot_cdf(
                self,
                moments,
                pointinterval,
                interval,
                levels,
                support,
                legend,
                color,
                alpha,
                figsize,
                ax,
            )
        else:
            return None

    def plot_ppf(
        self,
        moments=None,
        pointinterval=False,
        interval=None,
        levels=None,
        legend="legend",
        color=None,
        alpha=1,
        figsize=None,
        ax=None,
    ):
        """
        Plot the quantile function.

        Parameters
        ----------
        moments : str
            Compute moments. Use any combination of the strings ``m``, ``d``, ``v``, ``s`` or ``k``
            for the mean (μ), standard deviation (σ), variance (σ²), skew (γ) or kurtosis (κ)
            respectively. Other strings will be ignored. Defaults to None.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"` (default),
            equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
            Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"legend"`` a
            ``"title"`` or not include them ``None``.
        color : str
            Valid matplotlib color. Defaults to None, colors will be selected from the active
            color cycle.
        alpha : float
            Transparency of the line. Defaults to 1 (no transparency).
        figsize : tuple
            Size of the figure
        ax : matplotlib axes
        """
        if valid_scalar_params(self):
            return plot_ppf(
                self, moments, pointinterval, interval, levels, legend, color, alpha, figsize, ax
            )
        else:
            return None

    def plot_interactive(
        self,
        kind="pdf",
        xy_lim="both",
        pointinterval=True,
        interval=None,
        levels=None,
        figsize=None,
    ):
        """
        Interactive exploration of distributions parameters

        Parameters
        ----------
        kind : str:
            Type of plot. Available options are `pdf`, `cdf` and `ppf`.
        xy_lim : str or tuple
            Set the limits of the x-axis and/or y-axis.
            Defaults to `"both"`, the limits of both axis are fixed.
            Use `"auto"` for automatic rescaling of x-axis and y-axis.
            Or set them manually by passing a tuple of 4 elements,
            the first two for x-axis, the last two for y-axis. The tuple can have `None`.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"` (default),
            equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
            Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        figsize : tuple
            Size of the figure
        """
        check_inside_notebook()

        if valid_scalar_params(self, check_frozen=False):
            args = dict(zip(self.param_names, self.params))
        else:
            name = self.__class__.__name__
            if name in ["Truncated", "Censored"]:
                vals = copy(init_vals["Truncated"])
                vals.update(init_vals[self.dist.__class__.__name__])
                self._parametrization(**vals)
            else:
                args = init_vals[self.__class__.__name__]
                self._parametrization(**args)

        if xy_lim == "both":
            xlim = self._finite_endpoints("full")
            xvals = self.xvals("restricted")
            if kind == "pdf":
                max_pdf = np.max(self.pdf(xvals))
                ylim = (-max_pdf * 0.075, max_pdf * 1.5)
            elif kind == "ppf":
                max_ppf = self.ppf(0.999)
                ylim = (-max_ppf * 0.075, max_ppf * 1.5)
        elif isinstance(xy_lim, tuple):
            xlim = xy_lim[:2]
            ylim = xy_lim[2:]

        sliders = {}
        if self.__class__.__name__ == "Categorical":
            pointinterval = False
            name = self.param_names[0]
            params = self.params[0]
            names = [f"{name}_{i}" for i in range(len(params))]
            supports = [(0, 1) for _ in range(len(params))]

            for name, value, support in zip(names, params, supports):
                sliders[name] = get_slider(name, value, *support)
        else:
            for name, value, support in zip(self.param_names, self.params, self.params_support):
                sliders[name] = get_slider(name, value, *support)

        def plot(**args):  # pylint: disable=inconsistent-return-statements
            if self.__class__.__name__ == "Categorical":
                values = list(args.values())
                args = {list(args.keys())[0].split("_")[0]: values}
                if np.sum(values) > 1:
                    return None

            self._parametrization(**args)

            if kind == "pdf":
                ax = self.plot_pdf(
                    legend=False,
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    figsize=figsize,
                )
            elif kind == "cdf":
                ax = self.plot_cdf(
                    legend=False,
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    figsize=figsize,
                )
            elif kind == "ppf":
                ax = self.plot_ppf(
                    legend=False,
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    figsize=figsize,
                )
            if xy_lim != "auto" and kind != "ppf":
                ax.set_xlim(*xlim)
            if xy_lim != "auto" and kind != "cdf":
                ax.set_ylim(*ylim)

        return interactive(plot, **sliders)


class Continuous(Distribution):
    """Base class for continuous distributions."""

    def __init__(self):
        super().__init__()
        self.kind = "continuous"


class Discrete(Distribution):
    """Base class for discrete distributions."""

    def __init__(self):
        super().__init__()
        self.kind = "discrete"


class DistributionTransformer(Distribution):
    """Base class for distributions that transform other distributions"""

    def __init__(self):
        super().__init__()
        if not isinstance(self.dist, list):
            self.kind = self.dist.kind


def _continuous_xvals(lower_ep, upper_ep, n_points):
    return np.linspace(lower_ep, upper_ep, n_points)


def _discrete_xvals(lower_ep, upper_ep, n_points):
    upper_ep = int(upper_ep)
    lower_ep = int(lower_ep)
    range_x = upper_ep - lower_ep
    if range_x <= n_points:
        x_vals = np.arange(lower_ep, upper_ep + 1, dtype=int)
    else:
        x_vals = np.linspace(lower_ep, upper_ep + 1, n_points, dtype=int)

    return x_vals



import pandas as pd
import pymc as pm
import bambi as bmb
from preliz.distributions import Normal, LogNormal, Gamma
from preliz.ppls.agnostic import posterior_to_prior


SEED = 2945

data = Normal(0, 1).rvs(200, random_state=SEED)

with pm.Model() as model:
    a = pm.Normal("a", mu=0, sigma=1)
    b = pm.HalfNormal("b", sigma=[1, 1], shape=2)
    y = pm.Normal("y", mu=a, sigma=b[0], observed=data)  # pylint:disable = unsubscriptable-object
    idata = pm.sample(tune=200, draws=500, random_seed=SEED)


def test_p2p_pymc():
    posterior_to_prior(model, idata)
    assert 'Gamma\x1b[0m("b", alpha=' in posterior_to_prior(model, idata, new_families="auto")
    posterior_to_prior(model, idata, new_families=[LogNormal()])
    assert 'Gamma\x1b[0m("b", mu=' in posterior_to_prior(
        model, idata, new_families={"b": [Gamma(mu=0)]}
    )


bmb_data = pd.DataFrame(
    {
        "y": Normal(0, 1).rvs(117, random_state=SEED + 1),
        "x": Normal(0, 1).rvs(117, random_state=SEED + 2),
        "x1": Normal(0, 1).rvs(117, random_state=SEED + 3),
    }
)
bmb_prior = {"Intercept": bmb.Prior("Normal", mu=0, sigma=1)}
bmb_model = bmb.Model("y ~ x + x1", bmb_data, priors=bmb_prior)
bmb_idata = bmb_model.fit(tune=200, draws=200, random_seed=SEED)


def test_p2p_bambi():
    posterior_to_prior(bmb_model, bmb_idata)
    assert 'Gamma\x1b[0m", alpha=' in posterior_to_prior(bmb_model, bmb_idata, new_families="auto")
    posterior_to_prior(bmb_model, bmb_idata, new_families=[LogNormal()])
    assert 'Normal\x1b[0m", mu=' in posterior_to_prior(
        bmb_model, bmb_idata, new_families={"Intercept": [Normal(mu=1, sigma=1)]}
    )



import numpy as np
import pandas as pd
import preliz as pz
import pymc as pm
import pytest
from numpy.testing import assert_allclose

np.random.seed(42)


@pytest.mark.parametrize(
    "params",
    [
        {
            "mu_x": 0,
            "sigma_x": 10,
            "sigma_z": 10,
            "target": pz.Normal(mu=174, sigma=20),
            "X": np.random.normal(0, 10, 120),
            "new_x": 174,
            "new_z": 2.9,
        },
        {
            "mu_x": [0, 1],
            "sigma_x": [10, 10],
            "sigma_z": 10,
            "target": pz.Normal(mu=174, sigma=20),
            "X": np.random.normal(0, 10, 120),
            "new_x": [174, 174],
            "new_z": 2.9,
        },
        {
            "mu_x": 0,
            "sigma_x": 10,
            "sigma_z": 10,
            "target": [(pz.Normal(mu=174, sigma=20), 0.5), (pz.Normal(mu=176, sigma=19.5), 0.5)],
            "X": np.random.normal(0, 10, 120),
            "new_x": 175,
            "new_z": 2.9,
        },
        {
            "mu_x": [0, 1],
            "sigma_x": 10,
            "sigma_z": 10,
            "target": [
                (pz.Normal(mu=174, sigma=20), 0.5),
                (pz.Normal(mu=176, sigma=19.5), 0.4),
                (pz.StudentT(mu=174, sigma=20, nu=3), 0.1),
            ],
            "X": np.random.normal(0, 10, 120),
            "new_x": [175, 175],
            "new_z": 3,
        },
    ],
)
def test_ppe(params):
    Y = np.zeros_like(params["X"])
    target = params.pop("target")
    with pm.Model() as model:
        x = pm.Normal("x", shape=1 if isinstance(params["mu_x"], int) else 2)
        z = pm.HalfNormal("z")
        x_idx = (
            x if np.asarray(params["mu_x"]).size == 1 else x[np.repeat(np.arange(2), Y.size / 2)]
        )
        pm.Normal("y", x_idx, z, observed=Y)

    new_prior = (
        pz.ppe(model, target, method="projective").replace("\x1b[1m", "").replace("\x1b[0m", "")
    )
    exec_context = {}
    exec(new_prior, globals(), exec_context)
    model = exec_context.get("model")
    initial = model.initial_point()
    assert_allclose(initial["x"], params["new_x"])
    assert_allclose(initial["z_log__"], params["new_z"], atol=0.1)



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import eps, to_precision, from_precision, all_not_none
from ..internal.special import erf, erfinv, mean_and_std, ppf_bounds_cont


class Normal(Continuous):
    r"""
    Normal distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \mu, \sigma) =
           \frac{1}{\sigma \sqrt{2\pi}}
           \exp\left\{ -\frac{1}{2} \left(\frac{x-\mu}{\sigma}\right)^2 \right\}

    .. plot::
        :context: close-figs


        from preliz import Normal, style
        style.use('preliz-doc')
        mus = [0., 0., -2.]
        sigmas = [1, 0.5, 1]
        for mu, sigma in zip(mus, sigmas):
            Normal(mu, sigma).plot_pdf()

    ========  ==========================================
    Support   :math:`x \in \mathbb{R}`
    Mean      :math:`\mu`
    Variance  :math:`\sigma^2`
    ========  ==========================================

    Normal distribution has 2 alternative parameterizations. In terms of mean and
    sigma (standard deviation), or mean and tau (precision).

    The link between the 2 alternatives is given by

    .. math::

        \tau = \frac{1}{\sigma^2}

    Parameters
    ----------
    mu : float
        Mean.
    sigma : float
        Standard deviation (sigma > 0).
    tau : float
        Precision (tau > 0).
    """

    def __init__(self, mu=None, sigma=None, tau=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, sigma, tau)

    def _parametrization(self, mu=None, sigma=None, tau=None):
        if all_not_none(sigma, tau):
            raise ValueError(
                "Incompatible parametrization. Either use mu and sigma, or mu and tau."
            )

        names = ("mu", "sigma")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))

        if tau is not None:
            self.tau = tau
            sigma = from_precision(tau)
            names = ("mu", "tau")

        self.mu = mu
        self.sigma = sigma
        self.param_names = names
        if all_not_none(mu, sigma):
            self._update(mu, sigma)

    def _update(self, mu, sigma):
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.tau = to_precision(sigma)

        if self.param_names[1] == "sigma":
            self.params = (self.mu, self.sigma)
        elif self.param_names[1] == "tau":
            self.params = (self.mu, self.tau)

        self.is_frozen = True

    def pdf(self, x):
        return nb_pdf(x, self.mu, self.sigma)

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.sigma)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma)

    def entropy(self):
        return nb_entropy(self.sigma)

    def mean(self):
        return self.mu

    def mode(self):
        return self.mu

    def median(self):
        return self.mu

    def var(self):
        return self.sigma**2

    def std(self):
        return self.sigma

    def skewness(self):
        return 0

    def kurtosis(self):
        return 0

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.normal(self.mu, self.sigma, size)

    def _fit_moments(self, mean, sigma):
        self._update(mean, sigma)

    def _fit_mle(self, sample):
        self._update(*nb_fit_mle(sample))


@nb.njit(cache=True)
def nb_cdf(x, mu, sigma):
    return 0.5 * (1 + erf((x - mu) / (sigma * 2**0.5)))


@nb.njit(cache=True)
def nb_ppf(q, mu, sigma):
    return ppf_bounds_cont(mu + sigma * 2**0.5 * erfinv(2 * q - 1), q, -np.inf, np.inf)


@nb.njit(cache=True)
def nb_pdf(x, mu, sigma):
    x = np.asarray(x)
    return 1 / np.sqrt(2 * np.pi * sigma**2) * np.exp(-0.5 * ((x - mu) / sigma) ** 2)


@nb.njit(cache=True)
def nb_entropy(sigma):
    return 0.5 * (np.log(2 * np.pi * np.e * sigma**2))


@nb.njit(cache=True)
def nb_fit_mle(sample):
    return mean_and_std(sample)


@nb.njit(cache=True)
def nb_logpdf(x, mu, sigma):
    return -np.log(sigma) - 0.5 * np.log(2 * np.pi) - 0.5 * ((x - mu) / sigma) ** 2


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma):
    return -(nb_logpdf(x, mu, sigma)).sum()



import sys
import pytest

from numpy.testing import assert_allclose

from preliz import quartile
from preliz.distributions import (
    AsymmetricLaplace,
    Beta,
    Cauchy,
    ChiSquared,
    ExGaussian,
    Exponential,
    Gamma,
    Gumbel,
    HalfCauchy,
    HalfNormal,
    HalfStudentT,
    InverseGamma,
    Kumaraswamy,
    Laplace,
    Logistic,
    LogLogistic,
    LogNormal,
    LogitNormal,
    Moyal,
    Normal,
    Pareto,
    Rice,
    SkewStudentT,
    StudentT,
    Triangular,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
    # Bernoulli, quartile is not useful for Bernoulli distribution as we only have two states
    BetaBinomial,
    DiscreteUniform,
    DiscreteWeibull,
    Geometric,
    HyperGeometric,
    NegativeBinomial,
    Poisson,
    ZeroInflatedBinomial,
    ZeroInflatedNegativeBinomial,
    ZeroInflatedPoisson,
)


@pytest.mark.parametrize(
    "distribution, q1, q2, q3, result",
    [
        (AsymmetricLaplace(), -1, 1, 3, (1.0, 1.0, 2.885)),
        (Beta(), 0.3, 0.5, 0.7, (1.528, 1.528)),
        (Cauchy(), -1, 0, 1, (0, 1)),
        (ChiSquared(), 2, 4, 5.5, (4.329)),
        (ExGaussian(), 8, 9, 10, (9, 1.482, 0)),
        (ExGaussian(mu=9), 8, 9, 10, (1.482, 0)),
        (Exponential(), 0.5, 1, 2.5, (0.611)),
        (Gamma(), 0.5, 1, 2.5, (0.894, 0.523)),
        (Gumbel(), 0.5, 1, 2.5, (0.751, 1.265)),
        (HalfCauchy(), 0.5, 1, 3, (1.105)),
        (HalfNormal(), 0.5, 1, 2, (1.613)),
        (HalfStudentT(), 0.5, 1, 2, (2.393, 1.311)),
        (InverseGamma(), 0.2, 0.3, 0.4, (3.881, 1.019)),
        (Kumaraswamy(), 0.2, 0.3, 0.4, (2.199, 9.598)),
        (Laplace(), -1, 0, 1, (0, 1.442)),
        (Logistic(), -1, 0, 1, (0, 0.910)),
        (LogLogistic(), 1, 1.5, 2, (1.454, 3.143)),
        (LogNormal(), 0.5, 1, 2, (0, 1.027)),
        (LogitNormal(), 0.3, 0.45, 0.6, (-0.212, 0.929)),
        (Moyal(), 0.5, 1, 2, (0.620, 0.567)),
        (Normal(), -1, 0, 1, (0, 1.482)),
        (Pareto(), 0.5, 1, 4, (0.541, 0.289)),
        (Rice(), 2, 4, 6, (0, 3.395)),
        (SkewStudentT(), 2, 4, 6, (4, 2.648, 1.663, 1.663)),
        pytest.param(
            StudentT(),
            -1,
            0,
            1,
            (84576.43, 0, 1.482),
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 8), reason="third party implementations details"
            ),
        ),
        (StudentT(nu=4), -1, 0, 1, (0, 1.350)),
        (Triangular(), 0, 1, 2, (-2.414, 1.0, 4.414)),
        (TruncatedNormal(), -1, 0, 1, (0, 1.482)),
        (Uniform(), -1, 0, 1, (-2, 2)),
        (VonMises(), -1, 0, 1, (0, 0.656)),
        (Wald(), 0.5, 1, 2, (1.698, 1.109)),
        (Weibull(), 0.5, 1, 2, (1.109, 1.456)),
        (BetaBinomial(), 3, 5, 7, (2.323, 1.949, 10.0)),
        (DiscreteUniform(), -2, 0, 2, (-5, 5)),
        (DiscreteWeibull(), 2, 6, 7, (0.951, 1.487)),
        (Geometric(), 2, 4, 6, (0.17)),
        (HyperGeometric(), 3, 4, 5, (50, 10, 20)),
        (NegativeBinomial(), 3, 5, 10, (7.283, 2.167)),
        (Poisson(), 4, 5, 6, (5.641)),
        (ZeroInflatedBinomial(), 1, 4, 7, (0.660, 10.947, 0.670)),
        (ZeroInflatedBinomial(psi=0.7), 2, 4, 6, (10.0, 0.571)),
        (ZeroInflatedNegativeBinomial(), 2, 4, 6, (0.87, 5.24, 17.49)),
        (ZeroInflatedNegativeBinomial(psi=0.9), 2, 4, 6, (5.16, 11.32)),
        (ZeroInflatedPoisson(), 4, 5, 6, (1, 5.641)),
        (ZeroInflatedPoisson(psi=0.8), 2, 4, 6, (5.475)),
    ],
)
def test_quartile(distribution, q1, q2, q3, result):
    quartile(distribution, q1, q2, q3)

    assert_allclose(distribution.opt.x, result, atol=0.01)



# pylint: disable=arguments-differ
import numpy as np

from preliz.distributions.distributions import DistributionTransformer
from preliz.internal.distribution_helper import all_not_none, num_skewness, num_kurtosis
from preliz.internal.optimization import find_ppf


class Mixture(DistributionTransformer):
    r"""
    Mixture distribution

    This is not a distribution per se, but a modifier of univariate distributions.

    Given a series of base distributions with probability density mass/function ($p_i$).
    The pdf/pmf of a mixture distribution is:

    .. math::

        f(x) = \sum_{i=1}^n \, w_i \, p_i(x)

    .. plot::
        :context: close-figs

        from preliz import Normal, Mixture, style
        style.use('preliz-doc')
        Mixture([Normal(0, 0.5), Normal(2, 0.5)], [0.2, 0.8]).plot_pdf()

    Parameters
    ----------
    dists: List of Univariate PreliZ distributions
        Components of the mixture. They should be all discrete or all continuous.
    weights: list of floats
        Weights must be larger or equal to 0 and their sum must be positive.
        If the weights do not sum up to 1, they will be normalized.
    """

    def __init__(self, dists, weights=None):
        self.dist = dists
        self.weights = None
        super().__init__()
        if all(dist.kind == "discrete" for dist in self.dist):
            self.kind = "discrete"
        elif all(dist.kind == "continuous" for dist in self.dist):
            self.kind = "continuous"
        else:
            raise ValueError("mixture of discrete and continuous distributions are not supported")

        self._parametrization(weights)

    def _parametrization(self, weights=None):
        self.params = []
        self.param_names = []
        self.params_support = []
        for dist in self.dist:
            if not hasattr(dist, "params"):
                self.params.extend([None] * len(dist.param_names))
            else:
                self.params.extend(dist.params)
            self.param_names.extend(dist.param_names)
            self.params_support.append(dist.params_support)

        self.params.append(weights)
        self.param_names.append("weights")
        self.params_support.append((0, 1))

        self.support = np.min([dist.support[0] for dist in self.dist]), np.max(
            [dist.support[1] for dist in self.dist]
        )
        self.weights = np.asarray(weights)
        self.weights = self.weights / np.sum(self.weights)

        if all_not_none(*self.params):
            self.is_frozen = True

    def pdf(self, x):
        return np.sum(
            [dist.pdf(x) * weight for dist, weight in zip(self.dist, self.weights)], axis=0
        )

    def cdf(self, x):
        return np.sum(
            [dist.cdf(x) * weight for dist, weight in zip(self.dist, self.weights)], axis=0
        )

    def ppf(self, q):
        return find_ppf(self, q)

    def logpdf(self, x):
        return np.sum(
            [dist.logpdf(x) * weight for dist, weight in zip(self.dist, self.weights)], axis=0
        )

    def _neg_logpdf(self, x):
        return -self.logpdf(x).sum()

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        if self.kind == "discrete":
            return -np.sum(np.exp(logpdf) * logpdf)
        else:
            return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def mean(self):
        return np.sum(
            [dist.mean() * weight for dist, weight in zip(self.dist, self.weights)], axis=0
        )

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return (
            np.sum(
                [
                    weight * (dist.var() + (dist.mean() ** 2))
                    for dist, weight in zip(self.dist, self.weights)
                ],
                axis=0,
            )
            - self.mean() ** 2
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return num_skewness(self)

    def kurtosis(self):
        return num_kurtosis(self)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        dist_idx = random_state.choice(len(self.dist), size=size, p=self.weights)
        samples = []
        for idx, dist in enumerate(self.dist):
            n_samples = np.sum(dist_idx == idx)
            if n_samples:
                samples.append(dist.rvs(n_samples, random_state=random_state))
        return np.concatenate(samples)

    def _fit_moments(self, mean, sigma):
        for dist in self.dist:
            dist._fit_moments(mean, sigma)
        self.is_frozen = True
        self.weights = np.ones(len(self.dist)) / len(self.dist)



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from ..internal.optimization import optimize_ml
from ..internal.special import ppf_bounds_cont
from ..internal.distribution_helper import all_not_none, eps
from .distributions import Continuous


class Cauchy(Continuous):
    r"""
    Cauchy Distribution

    The pdf of this distribution is

    .. math::

        f(x \mid \alpha, \beta) =
            \frac{1}{\pi \beta [1 + (\frac{x-\alpha}{\beta})^2]}

    .. plot::
        :context: close-figs


        from preliz import Cauchy, style
        style.use('preliz-doc')
        alphas = [0., 0., -2.]
        betas = [.5, 1., 1.]
        for alpha, beta in zip(alphas, betas):
            Cauchy(alpha, beta).plot_pdf(support=(-5,5))

    ========  ==============================================================
    Support   :math:`x \in \mathbb{R}`
    Mean      undefined
    Variance  undefined
    ========  ==============================================================

    Parameters
    ----------
    alpha : float
        Location parameter.
    beta : float
        Scale parameter > 0.
    """

    def __init__(self, alpha=None, beta=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(alpha, beta)

    def _parametrization(self, alpha=None, beta=None):
        self.alpha = alpha
        self.beta = beta
        self.param_names = ("alpha", "beta")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))
        self.params = (self.alpha, self.beta)
        if all_not_none(alpha, beta):
            self._update(alpha, beta)

    def _update(self, alpha, beta):
        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.params = (self.alpha, self.beta)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.alpha, self.beta))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.beta)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.beta, -np.inf, np.inf)

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta)

    def entropy(self):
        return nb_entropy(self.beta)

    def mean(self):
        return np.nan

    def mode(self):
        return self.alpha

    def median(self):
        return self.alpha

    def var(self):
        return np.nan

    def std(self):
        return np.nan

    def skewness(self):
        return np.nan

    def kurtosis(self):
        return np.nan

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        random_samples = random_state.uniform(0, 1, size)
        return nb_rvs(random_samples, self.alpha, self.beta)

    def _fit_moments(self, mean, sigma):
        self._update(mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, alpha, beta):
    return 1 / np.pi * np.arctan((x - alpha) / beta) + 0.5


@nb.njit(cache=True)
def nb_ppf(q, alpha, beta, lower, upper):
    x_val = alpha + beta * np.tan(np.pi * (q - 0.5))
    return ppf_bounds_cont(x_val, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(beta):
    return np.log(4 * np.pi * beta)


@nb.njit(cache=True)
def nb_logpdf(x, alpha, beta):
    return -np.log(np.pi) - np.log(beta) - np.log(1 + ((x - alpha) / beta) ** 2)


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta):
    return -(nb_logpdf(x, alpha, beta)).sum()


@nb.njit(cache=True)
def nb_rvs(random_samples, alpha, beta):
    return alpha + beta * np.tan(np.pi * (random_samples - 0.5))



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from ..internal.special import cdf_bounds, ppf_bounds_cont
from ..internal.optimization import optimize_ml
from ..internal.distribution_helper import eps
from .distributions import Continuous


class HalfCauchy(Continuous):
    r"""
    HalfCauchy Distribution

    The pdf of this distribution is

    .. math::

        f(x \mid \beta) =
            \frac{2}{\pi \beta [1 + (\frac{x}{\beta})^2]}

    .. plot::
        :context: close-figs


        from preliz import HalfCauchy, style
        style.use('preliz-doc')
        for beta in [.5, 1., 2.]:
            HalfCauchy(beta).plot_pdf(support=(0,5))

    ========  ==========================================
    Support   :math:`x \in [0, \infty)`
    Mean      undefined
    Variance  undefined
    ========  ==========================================

    Parameters
    ----------
    beta : float
        Scale parameter :math:`\beta` (``beta`` > 0)
    """

    def __init__(self, beta=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(beta)

    def _parametrization(self, beta=None):
        self.beta = beta
        self.params = (self.beta,)
        self.param_names = ("beta",)
        self.params_support = ((eps, np.inf),)
        if self.beta is not None:
            self._update(self.beta)

    def _update(self, beta):
        self.beta = np.float64(beta)
        self.params = (self.beta,)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.beta))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.beta, 0, np.inf)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.beta, 0, np.inf)

    def logpdf(self, x):
        return nb_logpdf(x, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.beta)

    def entropy(self):
        return nb_entropy(self.beta)

    def mean(self):
        return np.inf

    def mode(self):
        return np.zeros_like(self.beta)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return np.inf

    def std(self):
        return np.inf

    def skewness(self):
        return np.nan

    def kurtosis(self):
        return np.nan

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        random_samples = random_state.uniform(0, 1, size)
        return nb_rvs(random_samples, self.beta)

    def _fit_moments(self, mean, sigma):  # pylint: disable=unused-argument
        self._update(sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, beta, lower, upper):
    prob = 2 / np.pi * np.arctan(x / beta)
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, beta, lower, upper):
    x_val = beta * np.tan(np.pi / 2 * q)
    return ppf_bounds_cont(x_val, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(beta):
    return np.log(2 * beta * np.pi)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, beta):
    if x < 0:
        return -np.inf
    else:
        return np.log(2) - np.log(np.pi * beta) - np.log(1 + (x / beta) ** 2)


@nb.njit(cache=True)
def nb_neg_logpdf(x, beta):
    return (-nb_logpdf(x, beta)).sum()


@nb.njit(cache=True)
def nb_rvs(random_samples, beta):
    return beta * np.tan(np.pi / 2 * random_samples)



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from scipy.special import gammaincc, gammainccinv  # pylint: disable=no-name-in-module

from ..internal.distribution_helper import all_not_none, any_not_none, eps
from ..internal.special import gammaln, digamma, cdf_bounds, ppf_bounds_cont, xlogy
from ..internal.optimization import optimize_ml
from .distributions import Continuous


class InverseGamma(Continuous):
    r"""
    Inverse gamma distribution, the reciprocal of the gamma distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \alpha, \beta) =
           \frac{\beta^{\alpha}}{\Gamma(\alpha)} x^{-\alpha - 1}
           \exp\left(\frac{-\beta}{x}\right)

    .. plot::
        :context: close-figs

        
        from preliz import InverseGamma, style
        style.use('preliz-doc')
        alphas = [1., 2., 3.]
        betas = [1., 1., .5]
        for alpha, beta in zip(alphas, betas):
            InverseGamma(alpha, beta).plot_pdf(support=(0, 3))

    ========  ===============================
    Support   :math:`x \in (0, \infty)`
    Mean      :math:`\dfrac{\beta}{\alpha-1}` for :math:`\alpha > 1`
    Variance  :math:`\dfrac{\beta^2}{(\alpha-1)^2(\alpha - 2)}` for :math:`\alpha > 2`
    ========  ===============================

    Inverse gamma distribution has 2 alternative parameterization. In terms of alpha and
    beta or mu (mean) and sigma (standard deviation).

    The link between the 2 alternatives is given by

    .. math::

       \alpha &= \frac{\mu^2}{\sigma^2} + 2 \\
       \beta  &= \frac{\mu^3}{\sigma^2} + \mu

    Parameters
    ----------
    alpha : float
        Shape parameter (alpha > 0).
    beta : float
        Scale parameter (beta > 0).
    mu : float
        Mean (mu > 0).
    sigma : float
        Standard deviation (sigma > 0)
    """

    def __init__(self, alpha=None, beta=None, mu=None, sigma=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(alpha, beta, mu, sigma)

    def _parametrization(self, alpha=None, beta=None, mu=None, sigma=None):
        if any_not_none(alpha, beta) and any_not_none(mu, sigma):
            raise ValueError(
                "Incompatible parametrization. Either use alpha and beta or mu and sigma."
            )

        self.param_names = ("alpha", "beta")
        self.params_support = ((eps, np.inf), (eps, np.inf))

        if any_not_none(mu, sigma):
            self.mu = mu
            self.sigma = sigma
            self.param_names = ("mu", "sigma")
            if all_not_none(mu, sigma):
                alpha, beta = _from_mu_sigma(mu, sigma)

        self.alpha = alpha
        self.beta = beta
        if all_not_none(self.alpha, self.beta):
            self._update(self.alpha, self.beta)

    def _update(self, alpha, beta):
        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.mu = _to_mu(self.alpha, self.beta)
        self.sigma = _to_sigma(self.alpha, self.beta)

        if self.param_names[0] == "alpha":
            self.params = (self.alpha, self.beta)
        elif self.param_names[1] == "sigma":
            self.params = (self.mu, self.sigma)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.beta, 0, np.inf)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.beta, 0, np.inf)

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta)

    def entropy(self):
        return nb_entropy(self.alpha, self.beta)

    def mean(self):
        return np.where(self.alpha > 1, self.beta / (self.alpha - 1), np.inf)

    def mode(self):
        return self.beta / (self.alpha + 1)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return np.where(
            self.alpha > 2, self.beta**2 / ((self.alpha - 1) ** 2 * (self.alpha - 2)), np.inf
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return np.where(self.alpha > 3, 4 * (self.alpha - 2) ** 0.5 / (self.alpha - 3), np.nan)

    def kurtosis(self):
        return np.where(
            self.alpha > 4,
            6 * (5 * self.alpha - 11) / ((self.alpha - 3) * (self.alpha - 4)),
            np.nan,
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return 1 / random_state.gamma(self.alpha, 1 / self.beta, size)

    def _fit_moments(self, mean, sigma):
        alpha, beta = _from_mu_sigma(mean, sigma)
        self._update(alpha, beta)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


def nb_cdf(x, alpha, beta, lower, upper):
    prob = gammaincc(alpha, beta / x)
    return cdf_bounds(prob, x, lower, upper)


def nb_ppf(q, alpha, beta, lower, upper):
    x_val = beta / gammainccinv(alpha, q)
    return ppf_bounds_cont(x_val, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(alpha, beta):
    return alpha + gammaln(alpha) - (1 + alpha) * digamma(alpha) + np.log(beta)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, beta):
    if x <= 0:
        return -np.inf
    else:
        return xlogy(alpha, beta) - gammaln(alpha) - xlogy((alpha + 1), x) - beta / x


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta):
    return -(nb_logpdf(x, alpha, beta)).sum()


def _from_mu_sigma(mu, sigma):
    alpha = mu**2 / sigma**2 + 2
    beta = mu**3 / sigma**2 + mu
    return alpha, beta


@nb.vectorize(nopython=True, cache=True)
def _to_mu(alpha, beta):
    if alpha > 1:
        return beta / (alpha - 1)
    else:
        return np.nan


@nb.vectorize(nopython=True, cache=True)
def _to_sigma(alpha, beta):
    if alpha > 2:
        return beta / ((alpha - 1) * (alpha - 2) ** 0.5)
    else:
        return np.nan



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Discrete
from ..internal.distribution_helper import all_not_none, eps
from ..internal.optimization import optimize_ml
from ..internal.special import xlogx, expit, logit


class Categorical(Discrete):
    R"""
    Categorical distribution.

    The most general discrete distribution. The pmf of this distribution is

    .. math:: f(x \mid p) = p_x

    .. plot::
        :context: close-figs


        from preliz import Categorical, style
        style.use('preliz-doc')
        ps = [[0.1, 0.6, 0.3], [0.3, 0.1, 0.1, 0.5]]
        for p in ps:
            Categorical(p).plot_pdf()

    ========  ===================================
    Support   :math:`x \in \{0, 1, \ldots, |p|-1\}`
    ========  ===================================

    Parameters
    ----------
    p : array of floats
        p > 0 and the elements of p must sum to 1.
    logit_p : float
        Alternative log odds for the probability of success.
    """

    def __init__(self, p=None, logit_p=None):
        super().__init__()
        self._parametrization(p, logit_p)

    def _parametrization(self, p=None, logit_p=None):
        if all_not_none(p, logit_p):
            raise ValueError("Incompatible parametrization. Either use p or logit_p.")

        self.param_names = "p"
        self.params_support = ((eps, np.inf),)

        if logit_p is not None:
            p = self._from_logit_p(logit_p)
            self.param_names = ("logit_p",)

        self.p = p
        self.logit_p = logit_p
        if self.p is not None:
            self.support = (0, len(p) - 1)
            self._update(self.p)

    def _from_logit_p(self, logit_p):
        return expit(logit_p)

    def _to_logit_p(self, p):
        return logit(p)

    def _get_frozen(self):
        frozen = None
        if all_not_none(self.params):
            frozen = self.dist(n=1, p=self.p)
        return frozen

    def _update(self, p):
        self.p = np.array(p)
        self._n = len(p)
        self.logit_p = self._to_logit_p(self.p)

        if self.param_names[0] == "p":
            self.params = (self.p,)
        elif self.param_names[0] == "logit_p":
            self.params = (self.logit_p,)

        self.is_frozen = True

    def pdf(self, x):
        x = np.atleast_1d(x)
        return nb_pdf(x, self.p)

    def cdf(self, x):
        x = np.atleast_1d(x)
        return nb_cdf(x, self.p)

    def ppf(self, q):
        q = np.atleast_1d(q)
        return nb_ppf(q, self.p)

    def logpdf(self, x):
        x = np.atleast_1d(x)
        return nb_logpdf(x, self.p)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.p)

    def entropy(self):
        return nb_entropy(self.p)

    def mean(self):
        return NotImplemented

    def mode(self):
        return self.p.argmax()

    def median(self):
        return NotImplemented

    def var(self):
        return NotImplemented

    def std(self):
        return NotImplemented

    def skewness(self):
        return NotImplemented

    def kurtosis(self):
        return NotImplemented

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.choice(self.p, size)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


def nb_pdf(x, p):
    pmf = np.zeros_like(x, dtype=float)
    valid_categories = np.where((x >= 0) & (x < len(p)))[0]
    pmf[valid_categories] = p[x[valid_categories]]
    return pmf


def nb_cdf(x, p):
    x = np.asarray(x, dtype=int)
    cdf = np.ones_like(x, dtype=float)
    cdf[x < 0] = 0
    valid_categories = np.where((x >= 0) & (x < len(p)))[0]
    cdf[valid_categories] = np.cumsum(p)[x[valid_categories]]
    return cdf


def nb_ppf(q, p):
    cumsum = np.cumsum(p)
    return np.searchsorted(cumsum, q)


@nb.njit(cache=True)
def nb_entropy(p):
    return -np.sum(xlogx(p))


def nb_logpdf(x, p):
    log_pmf = np.full_like(x, -np.inf, dtype=float)
    valid_categories = np.where((x >= 0) & (x < len(p)))[0]
    log_pmf[valid_categories] = np.log(p[x[valid_categories]])
    return log_pmf


def nb_neg_logpdf(x, p):
    return -(nb_logpdf(x, p)).sum()



# pylint: disable=arguments-differ
import numpy as np

from preliz.distributions.distributions import DistributionTransformer
from preliz.internal.distribution_helper import all_not_none
from preliz.internal.special import xlogx, xprody
from preliz.distributions.truncated import Truncated


class Censored(DistributionTransformer):
    r"""
    Censored distribution

    This is not a distribution per se, but a modifier of univariate distributions.

    Given a base distribution with cumulative distribution function (CDF) and
    probability density mass/function (PDF). The pdf of a Censored distribution is:

    .. math::

        \begin{cases}
            0 & \text{for } x < \text{lower}, \\
            \text{CDF}(\text{lower}) & \text{for } x = \text{lower}, \\
            \text{PDF}(x) & \text{for } \text{lower} < x < \text{upper}, \\
            1-\text{CDF}(\text{upper}) & \text {for } x = \text{upper}, \\
            0 & \text{for } x > \text{upper},
        \end{cases}

    .. plot::
        :context: close-figs

        
        from preliz import Normal, Censored, style
        style.use('preliz-doc')
        Censored(Normal(0, 1), -1, 1).plot_pdf(support=(-4, 4))
        Normal(0, 1).plot_pdf(alpha=0.5)
        

    Parameters
    ----------
    dist: PreliZ distribution
        Univariate PreliZ distribution which will be censored.
    lower: float or int
        Lower (left) censoring point. Use np.inf for no censoring.
    upper: float or int
        Upper (right) censoring point. Use np.inf for no censoring.
    """

    def __init__(self, dist, lower=None, upper=None, **kwargs):
        self.dist = dist
        super().__init__()
        self._parametrization(lower, upper, **kwargs)

    def _parametrization(self, lower=None, upper=None, **kwargs):
        dist_params = []
        if not kwargs:
            if hasattr(self.dist, "params"):
                kwargs = dict(zip(self.dist.param_names, self.dist.params))
            else:
                kwargs = dict(zip(self.dist.param_names, [None] * len(self.dist.param_names)))

        for key, value in kwargs.items():
            dist_params.append(value)
            setattr(self, key, value)

        if upper is None:
            self.upper = np.inf
        else:
            self.upper = upper

        if lower is None:
            self.lower = -np.inf
        else:
            self.lower = lower

        self.params = (*dist_params, self.lower, self.upper)
        self.param_names = (*self.dist.param_names, "lower", "upper")
        if all_not_none(*dist_params):
            self.dist._parametrization(**kwargs)
            self.is_frozen = True

        self.support = (
            np.maximum(self.dist.support[0], self.lower),
            np.minimum(self.dist.support[1], self.upper),
        )
        self.lower, self.upper = self.support
        self.params_support = (*self.dist.params_support, self.dist.support, self.dist.support)

    def mean(self):
        if self.kind == "discrete":
            p_low = self.dist.cdf(self.lower - 1)
        else:
            p_low = self.dist.cdf(self.lower)
        p_up = 1 - self.dist.cdf(self.upper)
        p_int = 1 - (p_low + p_up)
        mean_trunc = Truncated(self.dist, self.lower, self.upper).mean()
        return xprody(self.lower, p_low) + mean_trunc * p_int + xprody(self.upper, p_up)

    def median(self):
        return np.clip(self.dist.median(), self.lower, self.upper)

    def var(self):
        mean = self.mean()
        if self.kind == "discrete":
            p_low = self.dist.cdf(self.lower - 1)
        else:
            p_low = self.dist.cdf(self.lower)
        p_up = 1 - self.dist.cdf(self.upper)
        p_int = 1 - (p_low + p_up)
        trunc = Truncated(self.dist, self.lower, self.upper)
        mean_trunc = trunc.mean()
        var_trunc = trunc.var() + (mean_trunc - mean) ** 2
        return (
            xprody((self.lower - mean) ** 2, p_low)
            + var_trunc * p_int
            + xprody((self.upper - mean) ** 2, p_up)
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        mean = self.mean()
        std = self.std()
        if self.kind == "discrete":
            p_low = self.dist.cdf(self.lower - 1)
        else:
            p_low = self.dist.cdf(self.lower)
        p_up = 1 - self.dist.cdf(self.upper)
        p_int = 1 - (p_low + p_up)
        trunc = Truncated(self.dist, self.lower, self.upper)
        mean_trunc = trunc.mean()
        std_trunc = trunc.std()

        skew_trunc = trunc.skewness() + ((mean_trunc - mean) / std_trunc) ** 3
        return (
            xprody(((self.lower - mean) / std) ** 3, p_low)
            + skew_trunc * p_int
            + xprody(((self.upper - mean) / std) ** 3, p_up)
        )

    def kurtosis(self):
        mean = self.mean()
        std = self.std()
        if self.kind == "discrete":
            p_low = self.dist.cdf(self.lower - 1)
        else:
            p_low = self.dist.cdf(self.lower)
        p_up = 1 - self.dist.cdf(self.upper)
        p_int = 1 - (p_low + p_up)
        trunc = Truncated(self.dist, self.lower, self.upper)
        mean_trunc = trunc.mean()
        std_trunc = trunc.std()

        kurt_trunc = trunc.kurtosis() + 3 + ((mean_trunc - mean) / std_trunc) ** 4
        return (
            xprody(((self.lower - mean) / std) ** 4, p_low)
            + kurt_trunc * p_int
            + xprody(((self.upper - mean) / std) ** 4, p_up)
        ) - 3

    def rvs(self, size=None, random_state=None):
        return np.clip(self.dist.rvs(size, random_state), self.lower, self.upper)

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        vals = self.dist.cdf(x)
        return np.where(x < self.lower, 0, np.where(x > self.upper, 1, vals))

    def ppf(self, q):
        return np.clip(self.dist.ppf(q), self.lower, self.upper)

    def logpdf(self, x):
        x = np.asarray(x)
        vals = self.dist.logpdf(x)
        vals = np.where((x < self.lower) | (x > self.upper), -np.inf, vals)
        with np.errstate(divide="ignore"):
            vals = np.where(x == self.lower, np.log(self.dist.cdf(self.lower)), vals)
            if self.kind == "discrete":
                vals = np.where(x == self.upper, np.log1p(-self.dist.cdf(self.upper - 1)), vals)
            else:
                vals = np.where(x == self.upper, np.log1p(-self.dist.cdf(self.upper)), vals)

        return vals

    def entropy(self):
        p_low_inc = self.dist.cdf(self.lower)
        p_up = 1 - self.dist.cdf(self.upper)
        if self.kind == "discrete":
            p_l = np.nan_to_num(self.dist.pdf(self.lower))
            p_u = np.nan_to_num(self.dist.pdf(self.upper))
            p_low = p_low_inc - p_l
            p_up_inc = p_up + p_u
            xlogx_pl = xlogx(p_low)
            xlogx_pu = xlogx(p_up)
        else:
            p_low = p_low_inc
            p_up_inc = p_up
            xlogx_pl = xlogx_pu = 0

        p_int = 1 - (p_low + p_up)
        entropy_bound = -(xlogx(p_low_inc) + xlogx(p_up_inc))
        trunc_ent = Truncated(self.dist, self.lower, self.upper).entropy()
        entropy_interval = trunc_ent * p_int - xlogx(p_int) + xlogx_pl + xlogx_pu
        return entropy_interval + entropy_bound

    def _neg_logpdf(self, x):
        return -self.logpdf(x).sum()

    def _fit_moments(self, mean, sigma):
        self.dist._fit_moments(mean, sigma)
        self._parametrization(**dict(zip(self.dist.param_names, self.dist.params)))



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np
from scipy.special import pdtr, pdtrik  # pylint: disable=no-name-in-module

from .distributions import Discrete
from ..internal.distribution_helper import eps, all_not_none
from ..internal.optimization import optimize_ml, optimize_moments
from ..internal.special import gammaln, xlogy, cdf_bounds, ppf_bounds_disc


class ZeroInflatedPoisson(Discrete):
    R"""
    Zero-inflated Poisson distribution.

    Often used to model the number of events occurring in a fixed period
    of time when the times at which events occur are independent.
    The pmf of this distribution is

    .. math::

        f(x \mid \psi, \mu) = \left\{ \begin{array}{l}
            (1-\psi) + \psi e^{-\mu}, \text{if } x = 0 \\
            \psi \frac{e^{-\mu}\mu^x}{x!}, \text{if } x=1,2,3,\ldots
            \end{array} \right.

    .. plot::
        :context: close-figs

        
        from preliz import ZeroInflatedPoisson, style
        style.use('preliz-doc')
        psis = [0.7, 0.4]
        mus = [8, 4]
        for psi, mu in zip(psis, mus):
            ZeroInflatedPoisson(psi, mu).plot_pdf()

    ========  ================================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\psi \mu`
    Variance  :math:`\psi \mu (1+(1-\psi) \mu`
    ========  ================================

    Parameters
    ----------
    psi : float
        Expected proportion of Poisson variates (0 < psi < 1)
    mu : float
        Expected number of occurrences during the given interval
        (mu >= 0).
    """

    def __init__(self, psi=None, mu=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(psi, mu)

    def _parametrization(self, psi=None, mu=None):
        self.psi = psi
        self.mu = mu
        self.params = (self.psi, self.mu)
        self.param_names = ("psi", "mu")
        self.params_support = ((eps, 1 - eps), (eps, np.inf))
        if all_not_none(psi, mu):
            self._update(psi, mu)

    def _update(self, psi, mu):
        self.psi = np.float64(psi)
        self.mu = np.float64(mu)
        self.params = (self.psi, self.mu)
        self.is_frozen = True

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.psi, self.mu))

    def cdf(self, x):
        return nb_cdf(x, self.psi, self.mu, self.support[0], self.support[1])

    def ppf(self, q):
        return nb_ppf(q, self.psi, self.mu, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.psi, self.mu)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.psi, self.mu)

    def entropy(self):
        if self.mu < 50:
            x = self.xvals("full", 5000)
            logpdf = self.logpdf(x)
            return -np.sum(np.exp(logpdf) * logpdf)
        else:
            poisson_entropy = (
                0.5 * np.log(2 * np.pi * np.e * self.mu)
                - 1 / (12 * self.mu)
                - 1 / (24 * self.mu**2)
                - 19 / (360 * self.mu**3)
            )
            if self.psi == 1:
                return poisson_entropy
            else:
                # The var can be 0 with probability 1-psi or something else with probability psi
                zero_entropy = -(1 - self.psi) * np.logp(1 - self.psi) - self.psi * np.log(self.psi)
                # The total entropy is the weighted sum of the two entropies
                return (1 - self.psi) * zero_entropy + self.psi * poisson_entropy

    def mean(self):
        return self.psi * self.mu

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.psi * self.mu * (1 + (1 - self.psi) * self.mu)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return np.nan

    def kurtosis(self):
        return np.nan

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        zeros = random_state.uniform(size=size) > (1 - self.psi)
        poisson = random_state.poisson(self.mu, size=size)
        return zeros * poisson


# @nb.jit
# pdtr not supported by numba
def nb_cdf(x, psi, mu, lower, upper):
    p_prob = pdtr(x, mu)
    prob = (1 - psi) + psi * p_prob
    return cdf_bounds(prob, x, lower, upper)


# @nb.jit
# pdtr not supported by numba
def nb_ppf(q, psi, mu, lower, upper):
    q = np.asarray(q)
    vals = np.ceil(pdtrik(q, mu))
    vals1 = np.maximum(vals - 1, 0)
    temp = pdtr(vals1, mu)
    p_vals = np.where(temp >= q, vals1, vals)
    x_vals = (1 - psi) + psi * p_vals
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, psi, mu):
    if x < 0:
        return -np.inf
    elif x == 0:
        return np.log(np.exp(-mu) * psi - psi + 1)
    else:
        return np.log(psi) + xlogy(x, mu) - gammaln(x + 1) - mu


@nb.njit(cache=True)
def nb_neg_logpdf(x, psi, mu):
    return -(nb_logpdf(x, psi, mu)).sum()



"""Prior predictive check assistant."""

import ast
from random import shuffle
import warnings

try:
    import ipywidgets as widgets
except ImportError:
    pass
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial import KDTree


from preliz.internal.plot_helper import (
    check_inside_notebook,
    plot_pp_samples,
    plot_pp_mean,
)
from preliz.ppls.agnostic import get_prior_pp_samples, from_preliz
from preliz.ppls.bambi_io import from_bambi
from preliz.internal.predictive_helper import back_fitting_ppa, select_prior_samples
from preliz.distributions import Normal
from preliz.distributions.distributions import Distribution


def ppa(
    fmodel,
    draws=2000,
    references=0,
    boundaries=(-np.inf, np.inf),
    target=None,
    new_families=True,
    engine="preliz",
):
    """
    Prior predictive check assistant.

    This is an experimental method under development, use with caution.

    Parameters
    ----------
    model : PreliZ model
    draws : int
        Number of draws from the prior and prior predictive distribution
    references : int, float, list, tuple or dictionary
        Value(s) used as reference points representing prior knowledge. For example expected
        values or values that are considered extreme. Use a dictionary for labeled references.
    boundaries : tuple
        Hard boundaries (lower, upper). Posterior predictive samples with values outside these
        boundaries will be excluded from the analysis.
    target : tuple or PreliZ distribtuion
        Target distribution. The first shown distributions will be selected to be as close
        as possible to `target`. Available options are, a PreliZ distribution or a 2-tuple with
        the first element representing the mean and the second the standard deviation.
    new_families : bool
        If True, the method will return the best fitting distribution from a set of common
        distributions
    engine : str
        Library used to define the model. Either `preliz` or `bambi`. Defaults to `preliz`
    """
    check_inside_notebook(need_widget=True)

    warnings.warn(""""This is an experimental method under development, use with caution.""")

    filter_dists = FilterDistribution(
        fmodel, draws, references, boundaries, target, new_families, engine
    )
    filter_dists()

    output = widgets.Output()

    with output:
        references_widget = widgets.Text(
            value=str(references),
            placeholder="Int, Float, tuple or dict",
            description="references: ",
            disabled=False,
            layout=widgets.Layout(width="230px", margin="0 20px 0 0"),
        )
        button_carry_on = widgets.Button(description="carry on")
        button_return_prior = widgets.Button(description="return prior")
        radio_buttons_kind = widgets.RadioButtons(
            options=["pdf", "hist", "ecdf"],
            value="pdf",
            description=" ",
            disabled=False,
        )

        check_button_sharex = widgets.Checkbox(
            value=True, description="sharex", disabled=False, indent=False
        )

        button_carry_on.on_click(
            lambda event: filter_dists.carry_on(radio_buttons_kind.value, check_button_sharex.value)
        )

        button_return_prior.on_click(lambda event: filter_dists.on_return_prior())

        def kind_(_):
            kind = radio_buttons_kind.value
            try:
                filter_dists.references = ast.literal_eval(references_widget.value)
            except (ValueError, SyntaxError):
                filter_dists.references = None

            plot_pp_samples(
                filter_dists.pp_samples,
                filter_dists.display_pp_idxs,
                filter_dists.references,
                kind,
                check_button_sharex.value,
                filter_dists.fig,
            )
            plot_pp_mean(
                filter_dists.pp_samples,
                list(filter_dists.selected),
                filter_dists.references,
                kind,
                filter_dists.fig_pp_mean,
            )

        references_widget.observe(kind_, names=["value"])

        radio_buttons_kind.observe(kind_, names=["value"])

        check_button_sharex.observe(kind_, names=["value"])

        def click(event):
            if event.inaxes is not None:
                if event.inaxes not in filter_dists.clicked:
                    filter_dists.clicked.append(event.inaxes)
                else:
                    filter_dists.clicked.remove(event.inaxes)
                    plt.setp(event.inaxes.spines.values(), color="k", lw=1)

                for ax in filter_dists.clicked:
                    plt.setp(ax.spines.values(), color="C1", lw=3)
                filter_dists.fig.canvas.draw()

        filter_dists.fig.canvas.mpl_connect("button_press_event", click)

    controls = widgets.VBox([button_carry_on, button_return_prior])
    plot_combine = widgets.VBox([radio_buttons_kind, check_button_sharex])

    display(  # pylint:disable=undefined-variable
        widgets.HBox([references_widget, plot_combine, controls, output])
    )


class FilterDistribution:  # pylint:disable=too-many-instance-attributes
    def __init__(self, fmodel, draws, references, boundaries, target, new_families, engine):
        self.fmodel = fmodel
        self.source = ""  # string representation of the model
        self.draws = draws
        self.references = references
        self.boundaries = boundaries
        self.target = target
        self.new_families = new_families
        self.engine = engine
        self.pp_samples = None  # prior predictive samples
        self.prior_samples = None  # prior samples used for backfitting
        self.display_pp_idxs = None  # indices of the pp_samples to be displayed
        self.pp_octiles = None  # octiles computed from pp_samples
        self.ref_octiles = None  # octiles computed from the target distribution
        self.kdt = None  # KDTree used to find similar distributions
        self.model = None  # parsed model used for backfitting
        self.clicked = []  # axes clicked by the user
        self.choices = []  # indices of the pp_samples selected by the user and not yet used to
        # find new distributions, this list can increase or decrease in size
        self.selected = set()  # indices of all the pp_samples selected by the user + machine
        # this set can only increase in size
        self.distances = {}  # distances between as selected distribution and its nearest neighbor
        self.shown = set()  # keep track to avoid showing the same distribution twice.
        self.fig = None  # figure used to display the pp_samples
        self.fig_pp_mean = None  # figure used to display the mean of the pp_samples
        self.axes = None  # axes used to display the pp_samples

    def __call__(self):

        if self.engine == "preliz":
            variables, self.model = from_preliz(self.fmodel)
        elif self.engine == "bambi":
            self.fmodel, variables, self.model = from_bambi(self.fmodel, self.draws)

        print(variables, self.model)
        self.pp_samples, self.prior_samples = get_prior_pp_samples(
            self.fmodel, variables, self.draws, self.engine
        )

        if self.target is not None:
            self.add_target_dist()

        self.pp_octiles, self.kdt = self.compute_octiles()
        self.display_pp_idxs = self.initialize_subsamples(self.target)
        self.fig, self.axes = plot_pp_samples(
            self.pp_samples,
            self.display_pp_idxs,
            self.references,
        )
        self.fig_pp_mean = plot_pp_mean(self.pp_samples, self.selected, self.references)

    def add_target_dist(self):
        if isinstance(self.target, tuple):
            ref_sample = Normal(*self.target).rvs(self.pp_samples.shape[1])
        elif isinstance(self.target, Distribution):
            ref_sample = self.target.rvs(self.pp_samples.shape[1])

        self.ref_octiles = np.quantile(ref_sample, [0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875])

    def compute_octiles(self):
        """
        Compute the octiles for the prior predictive samples. This is used to find
        similar distributions using a KDTree.

        We have empirically found that octiles are a good choice, but this could
        be the consequence of limited testing.
        """
        pp_octiles = np.quantile(
            self.pp_samples, [0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875], axis=1
        ).T
        kdt = KDTree(pp_octiles)
        return pp_octiles, kdt

    def initialize_subsamples(self, target):
        """
        Initialize the subsamples to be displayed.

        If `target` is None, we search for the farthest_neighbor (this increases diversity)
        otherwise we search for the nearest_neighbor of target

        The initialization takes into account the boundaries provided by the user.

        Updates the `shown` set (already shown pp_samples) and if `target` is None
        also updates the  `distances` dictionary.
        """
        pp_idxs_to_display = []

        shown_list = []
        for idx, sample in enumerate(self.pp_samples):
            if np.min(sample) < self.boundaries[0] or np.max(sample) > self.boundaries[1]:
                shown_list.append(idx)

        self.shown.update(shown_list)

        # If we have not seen all the samples yet, we collect more
        if len(self.shown) != self.draws:
            if target is None:
                new = np.random.choice(list(set(range(0, len(self.pp_octiles))) - self.shown))
                pp_idxs_to_display.append(new)

                for _ in range(8):
                    farthest_neighbor = self.draws
                    while new in pp_idxs_to_display or new in self.shown:
                        _, new = self.kdt.query(
                            self.pp_octiles[pp_idxs_to_display[-1]], [farthest_neighbor], workers=-1
                        )
                        new = new.item()
                        farthest_neighbor -= 1
                    # Missing neighbors are indicated with index==sample_size
                    if new != self.draws:
                        pp_idxs_to_display.append(new)
            else:
                new = -1
                pp_idxs_to_display.append(new)

                for _ in range(9):
                    nearest_neighbor = 2
                    while new in pp_idxs_to_display:
                        distance, new = self.kdt.query(
                            self.ref_octiles, [nearest_neighbor], workers=-1
                        )
                        new = new.item()
                        nearest_neighbor += 1

                    if new != self.draws:
                        pp_idxs_to_display.append(new)
                        self.distances[new] = distance.item()

                pp_idxs_to_display = pp_idxs_to_display[1:]

            self.shown.update(pp_idxs_to_display)

        return pp_idxs_to_display

    def keep_sampling(self):
        """
        Find distribution similar to the ones in `choices`, but not already shown.
        If `choices` is empty return an empty selection.
        """
        if self.choices:
            new = self.choices.pop(0)
            samples = [new]

            for _ in range(9):
                nearest_neighbor = 2
                while new in samples or new in self.shown:
                    distance, new = self.kdt.query(
                        self.pp_octiles[samples[-1]], [nearest_neighbor], workers=-1
                    )
                    new = new.item()
                    nearest_neighbor += 1

                # Missing neighbors are indicated with index==self.draws
                if new != self.draws:
                    self.distances[new] = distance.item()
                    samples.append(new)

            self.shown.update(samples[1:])

            return samples[1:]
        else:
            return []

    def collect_more_samples(self):
        """
        Automatically extend the set of user selected distributions

        If the user has selected at least two distributions we automatically extend the selection
        by adding all the distributions that are close to the selected ones. To do so we use
        compute a max_dist, which is the trimmed mean of the distances between the already selected
        distributions. The trimmed mean is computed by discarding the lower and upper 10%.
        """
        selected_distances = np.array([v for k, v in self.distances.items() if k in self.selected])

        if len(selected_distances) > 2:
            q_r = np.quantile(selected_distances, [0.1, 0.9])
            max_dist = np.mean(
                selected_distances[(selected_distances > q_r[0]) & (selected_distances < q_r[1])]
            )
            upper = self.draws
        else:
            max_dist = np.inf
            upper = 3

        _, new = self.kdt.query(
            self.pp_octiles[list(self.selected)],
            range(2, upper),
            distance_upper_bound=max_dist,
            workers=-1,
        )
        new = new[new < self.draws].tolist()

        if np.any(np.isfinite(self.boundaries)):
            new_ = []
            for n_s in new:
                sample = self.pp_samples[n_s]
                if np.min(sample) > self.boundaries[0] and np.max(sample) < self.boundaries[1]:
                    new_.append(n_s)
            new = new_

        if new:
            self.selected.update(new)
            self.shown.update(new)

    def carry_on(self, kind, sharex):
        """
        Collect user input and update the plot.
        """
        self.fig.suptitle("")

        if self.clicked:
            self.choices.extend([int(ax.get_title()) for ax in self.clicked])
            shuffle(self.choices)
            self.selected.update(self.choices)
            self.collect_more_samples()

            for ax in self.clicked:
                plt.setp(ax.spines.values(), color="k", lw=1)
            for ax in self.axes:
                ax.cla()
            for ax in list(self.clicked):
                self.clicked.remove(ax)

        self.display_pp_idxs = self.keep_sampling()

        # if there is nothing to show initialize a new set of samples
        if not self.display_pp_idxs:
            self.display_pp_idxs = self.initialize_subsamples(None)

        plot_pp_mean(self.pp_samples, list(self.selected), self.references, kind, self.fig_pp_mean)

        if self.display_pp_idxs:
            plot_pp_samples(
                self.pp_samples, self.display_pp_idxs, self.references, kind, sharex, self.fig
            )
        else:
            # Instead of showing this message, we should resample.
            self.fig.clf()
            self.fig.suptitle("We have seen all the samples", fontsize=16)
            self.fig.canvas.draw()

    def on_return_prior(self):

        selected = list(self.selected)

        if len(selected) > 4:
            subsample = select_prior_samples(selected, self.prior_samples, self.model)

            string, _ = back_fitting_ppa(self.model, subsample, new_families=self.new_families)

            self.fig.clf()
            plt.text(0.05, 0.5, string, fontsize=14)

            plt.yticks([])
            plt.xticks([])
        else:
            self.fig.suptitle("Please select more distributions", fontsize=16)

        self.fig.canvas.draw()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np
from scipy.special import nbdtrik  # pylint: disable=no-name-in-module

from .distributions import Discrete
from ..internal.distribution_helper import eps, any_not_none, all_not_none
from ..internal.optimization import optimize_moments, optimize_ml
from ..internal.special import betainc, gammaln, xlogy, cdf_bounds, ppf_bounds_disc


class NegativeBinomial(Discrete):
    R"""
    Negative binomial distribution.

    The negative binomial distribution describes a Poisson random variable
    whose rate parameter is gamma distributed.
    Its pmf, parametrized by the parameters alpha and mu of the gamma distribution, is

    .. math::

       f(x \mid \mu, \alpha) =
           \binom{x + \alpha - 1}{x}
           (\alpha/(\mu+\alpha))^\alpha (\mu/(\mu+\alpha))^x

    .. plot::
        :context: close-figs

        
        from preliz import NegativeBinomial, style
        style.use('preliz-doc')
        mus = [1, 2, 8]
        alphas = [0.9, 2, 4]
        for mu, alpha in zip(mus, alphas):
            NegativeBinomial(mu, alpha).plot_pdf(support=(0, 20))

    ========  ==========================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\mu`
    Variance  :math:`\frac{\mu (\alpha + \mu)}{\alpha}`
    ========  ==========================

    The negative binomial distribution can be parametrized either in terms of mu and alpha,
    or in terms of n and p. The link between the parametrizations is given by

    .. math::

        p &= \frac{\alpha}{\mu + \alpha} \\
        n &= \alpha

    If it is parametrized in terms of n and p, the negative binomial describes the probability
    to have x failures before the n-th success, given the probability p of success in each trial.
    Its pmf is

    .. math::

        f(x \mid n, p) =
           \binom{x + n - 1}{x}
           (p)^n (1 - p)^x

    Parameters
    ----------
    alpha : float
        Gamma distribution shape parameter (alpha > 0).
    mu : float
        Gamma distribution mean (mu > 0).
    p : float
        Probability of success in each trial (0 < p < 1).
    n : float
        Number of target success trials (n > 0)
    """

    def __init__(self, mu=None, alpha=None, p=None, n=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(mu, alpha, p, n)

    def _parametrization(self, mu=None, alpha=None, p=None, n=None):
        if any_not_none(mu, alpha) and any_not_none(p, n):
            raise ValueError("Incompatible parametrization. Either use mu and alpha, or p and n.")

        self.param_names = ("mu", "alpha")
        self.params_support = ((eps, np.inf), (eps, np.inf))

        if any_not_none(p, n):
            self.p = p
            self.n = n
            self.param_names = ("p", "n")
            if all_not_none(p, n):
                mu, alpha = self._from_n_p(n, p)

        self.mu = mu
        self.alpha = alpha
        if all_not_none(mu, alpha):
            self._update(mu, alpha)

    def _from_n_p(self, n, p):
        mu = n * (1 / p - 1)
        return mu, n

    def _to_n_p(self, mu, alpha):
        p = alpha / (mu + alpha)
        return alpha, p

    def _update(self, mu, alpha):
        self.mu = np.float64(mu)
        self.alpha = np.float64(alpha)
        self.n, self.p = self._to_n_p(self.mu, self.alpha)

        if self.param_names[0] == "mu":
            self.params = (self.mu, self.alpha)
        elif self.param_names[0] == "p":
            self.params = (self.p, self.n)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.n, self.p))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.n, self.p, self.support[0], self.support[1])

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.n, self.p, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.n, self.p)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.n, self.p)

    def entropy(self):
        x = self.xvals("full", 5000)
        logpdf = self.logpdf(x)
        return -np.sum(np.exp(logpdf) * logpdf)

    def mean(self):
        return self.mu

    def mode(self):
        return np.where(self.n < 1, 0, np.floor((self.n - 1) * (1 - self.p) / self.p))

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.mu**2 / self.alpha + self.mu

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (2 - self.p) / ((1 - self.p) * self.n) ** 0.5

    def kurtosis(self):
        return 6 / self.n + self.p**2 / ((1 - self.p) * self.n)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.negative_binomial(self.n, self.p, size=size)

    def _fit_moments(self, mean, sigma=None):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, n, p, lower, upper):
    prob = betainc(n, x + 1, p)
    return cdf_bounds(prob, x, lower, upper)


# @nb.jit
# bdtrik not supported by numba
def nb_ppf(q, n, p, lower, upper):
    x_vals = np.ceil(nbdtrik(q, n, p))
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(y, n, p):
    if y < 0:
        return -np.inf
    else:
        return gammaln(y + n) - gammaln(n) - gammaln(y + 1) + xlogy(n, p) + xlogy(y, 1 - p)


@nb.njit(cache=True)
def nb_neg_logpdf(y, n, p):
    return -(nb_logpdf(y, n, p)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np
from scipy.special import pdtr, pdtrik  # pylint: disable=no-name-in-module

from .distributions import Discrete
from ..internal.distribution_helper import eps
from ..internal.special import gammaln, xlogy, cdf_bounds, ppf_bounds_disc


class Poisson(Discrete):
    R"""
    Poisson distribution.

    Often used to model the number of events occurring in a fixed period
    of time when the times at which events occur are independent.
    The pmf of this distribution is

    .. math:: f(x \mid \mu) = \frac{e^{-\mu}\mu^x}{x!}

    .. plot::
        :context: close-figs


        from preliz import Poisson, style
        style.use('preliz-doc')
        for mu in [0.5, 3, 8]:
            Poisson(mu).plot_pdf()

    ========  ==========================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\mu`
    Variance  :math:`\mu`
    ========  ==========================

    Parameters
    ----------
    mu: float
        Expected number of occurrences during the given interval
        (mu >= 0).

    Notes
    -----
    The Poisson distribution can be derived as a limiting case of the
    binomial distribution.
    """

    def __init__(self, mu=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(mu)

    def _parametrization(self, mu=None):
        self.mu = mu
        self.params = (self.mu,)
        self.param_names = ("mu",)
        self.params_support = ((eps, np.inf),)
        if mu is not None:
            self._update(mu)

    def _update(self, mu):
        self.mu = np.float64(mu)
        self.params = (self.mu,)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.mu))

    def cdf(self, x):
        return nb_cdf(x, self.mu, self.support[0], self.support[1])

    def ppf(self, q):
        return nb_ppf(q, self.mu, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.mu)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu)

    def entropy(self):
        x = np.arange(0, self.ppf(0.9999) + 1, dtype=int)
        logpdf = self.logpdf(x)
        return -np.sum(np.exp(logpdf) * logpdf)

    def mean(self):
        return self.mu

    def mode(self):
        return np.floor(self.mu)

    def median(self):
        return np.floor(self.mu + 1 / 3 - 0.02 / self.mu)

    def var(self):
        return self.mu

    def std(self):
        return self.mu**0.5

    def skewness(self):
        return 1 / self.mu**0.5

    def kurtosis(self):
        return 1 / self.mu

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.poisson(self.mu, size=size)

    def _fit_moments(self, mean, sigma=None):  # pylint: disable=unused-argument
        self._update(mean)

    def _fit_mle(self, sample):
        self._update(nb_fit_mle(sample))


# @nb.jit
# pdtr not supported by numba
def nb_cdf(x, mu, lower, upper):
    prob = pdtr(x, mu)
    return cdf_bounds(prob, x, lower, upper)


# @nb.jit
# pdtr not supported by numba
def nb_ppf(q, mu, lower, upper):
    q = np.asarray(q)
    vals = np.ceil(pdtrik(q, mu))
    vals1 = np.maximum(vals - 1, 0)
    temp = pdtr(vals1, mu)
    x_vals = np.where(temp >= q, vals1, vals)
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.njit(cache=True)
def nb_fit_mle(sample):
    return np.mean(sample)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu):
    if x < 0:
        return -np.inf
    else:
        return xlogy(x, mu) - gammaln(x + 1) - mu


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu):
    return -(nb_logpdf(x, mu)).sum()



try:
    import ipywidgets as widgets
except ImportError:
    pass

from preliz.internal.optimization import fit_to_quartile
from preliz.internal.plot_helper import (
    create_figure,
    check_inside_notebook,
    representations,
    reset_dist_panel,
)
from preliz.internal.distribution_helper import process_extra, get_distributions


class QuartileInt:
    """
    Prior elicitation for 1D distributions from quartiles.

    Parameters
    ----------
    q1 : float
        First quartile, i.e 0.25 of the mass is below this point.
    q2 : float
        Second quartile, i.e 0.50 of the mass is below this point. This is also know
        as the median.
    q3 : float
        Third quartile, i.e 0.75 of the mass is below this point.
    dist_names: list
        List of distributions names to be used in the elicitation. If None, almost all 1D
        distributions available in PreliZ will be used. Some distributions like Uniform or
        Cauchy are omitted by default.
    figsize: Optional[Tuple[int, int]]
        Figure size. If None it will be defined automatically.

    Note
    ----
    Use the `params` text box to parametrize distributions, for instance write
    `BetaScaled(lower=-1, upper=10)` to specify the upper and lower bounds of BetaScaled
    distribution. To parametrize more that one distribution use commas for example
    `StudentT(nu=3), TruncatedNormal(lower=-2, upper=inf)`

    References
    ----------
    * Morris D.E. et al. (2014) see https://doi.org/10.1016/j.envsoft.2013.10.010
    * See quartile mode http://optics.eee.nottingham.ac.uk/match/uncertainty.php
    """

    def __init__(self, q1=1, q2=2, q3=3, dist_names=None, figsize=None):
        self._q1 = q1
        self._q2 = q2
        self._q3 = q3
        self.dist = None
        self._dist_names = dist_names
        self._figsize = figsize

        check_inside_notebook(need_widget=True)
        self._widgets = self._get_widgets()

        self._output = widgets.Output()

        with self._output:
            if self._figsize is None:
                self._figsize = (8, 6)

            self._fig, self._ax_fit = create_figure(self._figsize)
            self._setup_observers()

            self._fig.canvas.mpl_connect(
                "button_release_event",
                lambda event: self._match_distribution(),
            )

        self._match_distribution()
        controls = widgets.VBox(
            [
                self._widgets["w_q1"],
                self._widgets["w_q2"],
                self._widgets["w_q3"],
                self._widgets["w_extra"],
            ]
        )
        display(  # pylint:disable=undefined-variable
            widgets.HBox([controls, self._widgets["w_repr"], self._widgets["w_distributions"]])
        )

    def _get_widgets(self):
        width_entry_text = widgets.Layout(width="150px")
        width_repr_text = widgets.Layout(width="250px")
        width_distribution_text = widgets.Layout(width="150px", height="125px")

        w_q1 = widgets.FloatText(
            value=self._q1,
            step=0.1,
            description="q1",
            disabled=False,
            layout=width_entry_text,
        )

        w_q2 = widgets.FloatText(
            value=self._q2,
            step=0.1,
            description="q2",
            disabled=False,
            layout=width_entry_text,
        )

        w_q3 = widgets.FloatText(
            value=self._q3,
            step=0.1,
            description="q3",
            disabled=False,
            layout=width_entry_text,
        )

        w_extra = widgets.Textarea(
            value="",
            placeholder="Pass extra parameters",
            description="params:",
            disabled=False,
            layout=width_repr_text,
        )

        w_repr = widgets.RadioButtons(
            options=["pdf", "cdf", "ppf"],
            value="pdf",
            description="",
            disabled=False,
            layout=width_repr_text,
        )

        if self._dist_names is None:

            default_dist = ["Normal", "BetaScaled", "Gamma", "LogNormal", "StudentT"]

            self._dist_names = [
                "AsymmetricLaplace",
                "BetaScaled",
                "ChiSquared",
                "ExGaussian",
                "Exponential",
                "Gamma",
                "Gumbel",
                "HalfNormal",
                "HalfStudentT",
                "InverseGamma",
                "Laplace",
                "LogNormal",
                "Logistic",
                "LogitNormal",
                "Moyal",
                "Normal",
                "Pareto",
                "Rice",
                "SkewNormal",
                "StudentT",
                "Triangular",
                "VonMises",
                "Wald",
                "Weibull",
                "BetaBinomial",
                "DiscreteWeibull",
                "Geometric",
                "NegativeBinomial",
                "Poisson",
            ]
        else:
            default_dist = self._dist_names

        w_distributions = widgets.SelectMultiple(
            options=self._dist_names,
            value=default_dist,
            description="",
            disabled=False,
            layout=width_distribution_text,
        )

        return {
            "w_q1": w_q1,
            "w_q2": w_q2,
            "w_q3": w_q3,
            "w_extra": w_extra,
            "w_repr": w_repr,
            "w_distributions": w_distributions,
        }

    def _match_distribution(self):
        q1 = float(self._widgets["w_q1"].value)
        q2 = float(self._widgets["w_q2"].value)
        q3 = float(self._widgets["w_q3"].value)
        extra_pros = process_extra(self._widgets["w_extra"].value)
        fitted_dist = None

        if q1 < q2 < q3:
            reset_dist_panel(self._ax_fit, yticks=False)

            fitted_dist = fit_to_quartile(
                get_distributions(self._widgets["w_distributions"].value), q1, q2, q3, extra_pros
            )

            if fitted_dist is None:
                self._ax_fit.set_title("domain error")
            else:
                representations(fitted_dist, self._widgets["w_repr"].value, self._ax_fit)
        else:
            reset_dist_panel(self._ax_fit, yticks=True)
            self._ax_fit.set_title("quantiles must follow the order: q1 < q2 < q3 ")

        self._fig.canvas.draw()

        self.dist = fitted_dist

    def _setup_observers(self):
        def _match_distribution_(_):
            self._match_distribution()

        self._widgets["w_repr"].observe(_match_distribution_)
        self._widgets["w_distributions"].observe(_match_distribution_)
        self._widgets["w_q1"].observe(_match_distribution_)
        self._widgets["w_q2"].observe(_match_distribution_)
        self._widgets["w_q3"].observe(_match_distribution_)



import numpy as np

from preliz.internal.distribution_helper import get_distributions
from preliz.internal.distribution_helper import process_extra
from preliz.internal.optimization import fit_to_epdf


def combine_roulette(responses, weights=None, dist_names=None, params=None):
    """
    Combine multiple elicited distributions into a single distribution.

    Parameters
    ----------
    responses : list of tuples
        Typically, each tuple comes from the ``.inputs`` attribute of a ``Roulette`` object and
        represents a single elicited distribution.
    weights : array-like, optional
        Weights for each elicited distribution. Defaults to None, i.e. equal weights.
        The sum of the weights must be equal to 1, otherwise it will be normalized.
    dist_names: list
        List of distributions names to be used in the elicitation.
        Defaults to ["Normal", "BetaScaled", "Gamma", "LogNormal", "StudentT"].
    params : str, optional
        Extra parameters to be passed to the distributions. The format is a string with the
        PreliZ's distribution name followed by the argument to fix.
        For example: "TruncatedNormal(lower=0), StudentT(nu=8)".

    Returns
    -------
    PreliZ distribution
    """

    if params is not None:
        extra_pros = process_extra(params)
    else:
        extra_pros = []

    if weights is None:
        weights = np.full(len(responses), 1 / len(responses))
    else:
        weights = np.array(weights, dtype=float)

    if np.any(weights < 0):
        raise ValueError("The weights must be positive.")

    weights /= weights.sum()

    if not all(records[3:] == responses[0][3:] for records in responses):
        raise ValueError(
            "To combine single elicitation instances, the grid should be the same for all of them."
        )

    if dist_names is None:
        dist_names = ["Normal", "BetaScaled", "Gamma", "LogNormal", "StudentT"]

    new_pdf = {}
    for records, weight in zip(responses, weights):
        chips = records[2]
        for x_i, pdf_i in zip(records[0], records[1]):
            if x_i in new_pdf:
                new_pdf[x_i] += pdf_i * weight * chips
            else:
                new_pdf[x_i] = pdf_i * weight * chips

    total = sum(new_pdf.values())
    mean = 0
    for x_i, pdf_i in new_pdf.items():
        val = pdf_i / total
        mean += x_i * val
        new_pdf[x_i] = val

    var = 0
    for x_i, pdf_i in new_pdf.items():
        var += pdf_i * (x_i - mean) ** 2
    std = var**0.5

    # Assuming all the elicited distributions have the same x_min and x_max
    x_min = responses[0][3]
    x_max = responses[0][4]

    fitted_dist = fit_to_epdf(
        get_distributions(dist_names),
        list(new_pdf.keys()),
        list(new_pdf.values()),
        mean,
        std,
        x_min,
        x_max,
        extra_pros,
    )

    return fitted_dist



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
# pylint: disable=invalid-name
import numba as nb
import numpy as np
from .distributions import Discrete

from ..internal.special import betaln, cdf_bounds
from ..internal.optimization import optimize_ml, optimize_moments, find_ppf
from ..internal.distribution_helper import all_not_none

eps = np.finfo(float).eps


class HyperGeometric(Discrete):
    R"""
    Discrete hypergeometric distribution.

    The probability of :math:`x` successes in a sequence of :math:`n` bernoulli
    trials taken without replacement from a population of :math:`N` objects,
    containing :math:`k` good (or successful or Type I) objects.
    The pmf of this distribution is

    .. math:: f(x \mid N, n, k) = \frac{\binom{k}{x}\binom{N-k}{n-x}}{\binom{N}{n}}

    .. plot::
        :context: close-figs


        from preliz import HyperGeometric, style
        style.use('preliz-doc')
        N = 50
        k = 10
        for n in [20, 25]:
            HyperGeometric(N, k, n).plot_pdf(support=(1,15))

    ========  =============================
    Support   :math:`x \in \left[\max(0, n - N + k), \min(k, n)\right]`
    Mean      :math:`\dfrac{nk}{N}`
    Variance  :math:`\dfrac{(N-n)nk(N-k)}{(N-1)N^2}`
    ========  =============================

    Parameters
    ----------
    N : int
        Total size of the population (N > 0)
    k : int
        Number of successful individuals in the population (0 <= k <= N)
    n : int
        Number of samples drawn from the population (0 <= n <= N)
    """

    def __init__(self, N=None, k=None, n=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(N, k, n)

    def _parametrization(self, N=None, k=None, n=None):
        self.N = N
        self.k = k
        self.n = n
        self.param_names = ("N", "k", "n")
        self.params_support = ((eps, np.inf), (eps, self.N), (eps, self.N))
        if all_not_none(self.N, self.k, self.n):
            self._update(N, k, n)

    def _update(self, N, k, n):
        self.N = np.int64(N)
        self.k = np.int64(k)
        self.n = np.int64(n)
        self.params = (self.N, self.k, self.n)
        self.support = (max(0, n - N + k), min(k, n))
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        if isinstance(x, (np.ndarray, list, tuple)):
            cdf_values = np.zeros_like(x, dtype=float)
            for i, val in enumerate(x):
                x_vals = np.arange(self.support[0], val + 1)
                cdf_values[i] = np.sum(self.pdf(x_vals))
            return cdf_bounds(cdf_values, x, *self.support)
        else:
            x_vals = np.arange(self.support[0], x + 1)
            return cdf_bounds(np.sum(self.pdf(x_vals)), x, *self.support)

    def ppf(self, q):
        q = np.asarray(q)
        return find_ppf(self, q)

    def logpdf(self, x):
        return nb_logpdf(x, self.N, self.k, self.n, *self.support)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.N, self.k, self.n, *self.support)

    def entropy(self):
        x_values = self.xvals("full")
        logpdf = self.logpdf(x_values)
        return -np.sum(np.exp(logpdf) * logpdf)

    def mean(self):
        return self.n * self.k / self.N

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return (
            self.n * self.k / self.N * (self.N - self.k) / self.N * (self.N - self.n) / (self.N - 1)
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        numerator = (self.N - 2 * self.k) * (self.N - 1) ** 0.5 * (self.N - 2 * self.n)
        denominator = (self.n * self.k * (self.N - self.k) * (self.N - self.n)) ** 0.5 * (
            self.N - 2
        )
        return numerator / denominator

    def kurtosis(self):
        return (
            1
            / (
                self.n
                * self.k
                * (self.N - self.k)
                * (self.N - self.n)
                * (self.N - 2)
                * (self.N - 3)
            )
            * (
                (self.N - 1)
                * self.N**2
                * (
                    self.N * (self.N + 1)
                    - 6 * self.k * (self.N - self.k)
                    - 6 * self.n * (self.N - self.n)
                )
                + 6 * self.n * self.k * (self.N - self.k) * (self.N - self.n) * (5 * self.N - 6)
            )
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.hypergeometric(self.k, self.N - self.k, self.n, size=size)

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, N, k, n, lower, upper):
    if x < lower:
        return -np.inf
    if x > upper:
        return -np.inf
    else:
        good = k
        bad = N - k
        tot = good + bad
        result = (
            betaln(good + 1, 1)
            + betaln(bad + 1, 1)
            + betaln(tot - n + 1, n + 1)
            - betaln(x + 1, good - x + 1)
            - betaln(n - x + 1, bad - n + x + 1)
            - betaln(tot + 1, 1)
        )
        return result


@nb.njit(cache=True)
def nb_neg_logpdf(x, N, k, n, lower, upper):
    return -(nb_logpdf(x, N, k, n, lower, upper)).sum()
